import numpy as np
from easier.bernstein import BernsteinFitter


def test_polynomial_coefficients_match_predict():
    # Generate some test data
    np.random.seed(42)
    x = np.linspace(-5, 5, 100)
    y = np.sin(x) + 0.1 * np.random.randn(len(x))

    # Fit the Bernstein polynomial
    fitter = BernsteinFitter()
    fitter.fit(x, y, degree=10)

    # Get polynomial coefficients
    coeffs = fitter.get_polynomial_coefficients()

    # Generate test points
    x_test = np.linspace(-5, 5, 50)

    # Get predictions using both methods
    y_bernstein = fitter.predict(x_test)
    y_poly = np.polyval(coeffs, x_test)

    # Check that they match within numerical tolerance
    np.testing.assert_allclose(y_bernstein, y_poly, rtol=1e-10, atol=1e-10)



import base64
from textwrap import dedent


class examples:
    """
    A descriptor whose only purpose is to print help text
    """

    def __get__(self, *args, **kwargs):
        print(
            dedent(
                """
                from easier import Crypt
                message = 'attack at dawn'
                encrypted_blob = Crypt().encrypt(message, 'my_password')
                unecrypted_text = Crypt().decrypt(encrypted_blob, 'my_password')
                print(f'Original: {repr(message)}')
                print(f'Encryped Blob: {repr(encrypted_blob)}')
                print(f'Unencrypted Message: {repr(unecrypted_text)}')
            """
            )
        )
        return None


class Crypt:
    """
    This is a utility to encrypt a string to a hex-encoded binary blob
    The code was copied from
    https://cryptography.io/en/latest/fernet/#using-passwords-with-fernet
    """

    examples = examples()

    @property
    def _kdf(self):
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
        from cryptography.hazmat.primitives import hashes

        return PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=b":h\x8a\xff\xda~}Dx\xa8\x80Q\xf3\x92\x93\x06",
            iterations=100000,
            backend=default_backend(),
        )

    def encrypt(self, message_string, password_string):
        from cryptography.fernet import Fernet

        password = password_string.encode()
        message = message_string.encode()
        key = base64.urlsafe_b64encode(self._kdf.derive(password))
        f = Fernet(key)
        token = f.encrypt(message)
        token_str = token.hex()
        return token_str

    def decrypt(self, encrypted_hex_string, password_string):
        from cryptography.fernet import Fernet

        password = password_string.encode()
        key = base64.urlsafe_b64encode(self._kdf.derive(password))
        f = Fernet(key)
        token = bytes.fromhex(encrypted_hex_string)
        message = f.decrypt(token).decode()
        return message



import datetime
import glob
import os
import pickle
import sys
import traceback
import warnings
from copy import copy, deepcopy
from textwrap import dedent


def tqdm_flex(iterable):
    """
    Adds the appropriate tqdm wrapper around an iterable
    """
    import easier as ezr

    if ezr.in_notebook():
        try:
            import tqdm.notebook as tqdm

            return tqdm.tqdm(iterable)
        except Exception:
            return iterable
    else:
        try:
            import tqdm

            return tqdm.tqdm(iterable)
        except Exception:
            return iterable


def python_type():
    """
    A utility to determine if running under ipython or jupyter
    """
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return "jupyter"  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return "ipython"  # Terminal running IPython
        else:
            return "other"  # Other type (?)
    except NameError:
        return "other"  # Probably standard Python interpreter


def in_notebook():
    """
    Determine if running in notebook (see python_type)
    """
    return python_type() == "jupyter"


def django_reconnect():  # pragma: no cover
    """
    Fixes dropped postgres connection in jupyter notebooks.
    """
    from django.db import connections

    conn = connections["default"]
    conn.connect()


def mute_warnings():  # pragma: no cover
    """
    Mute all Python warnings
    """
    import warnings

    warnings.filterwarnings("ignore")


def screen_width_full():  # pragma: no cover
    from IPython.core.display import display, HTML

    display(HTML("<style>.container { width:100% !important; }</style>"))


def print_error(tag="", verbose=False, buffer=None):  # pragma: no cover
    """
    Function for printing errors in except block.
    Args:
        tag: Optional string to print after exception info
        verbose: Only print traceback when verbose = True
        buffer: The buffer to print to (default: sys.stdout)
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()

    if buffer is None:
        buffer = sys.stderr

    if verbose:
        traceback.print_tb(exc_traceback, limit=None, file=buffer)

    if tag:
        tag = f" :: {tag.strip()}"

    print(f"{exc_type.__name__}: {exc_value}{tag}", file=buffer)


class cached_property(object):
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    Clear the cache by just deleteing the property

    class Person:
        @cached_property
        def first_name(self):
            return 'Monte'

    p = Person()

    # Compute and return first name
    f = p.first_name

    # Accessed cached first name by calling again
    f = p.first_name

    # Clear the cache by deleting the property
    # Will raise attribute error if cache is empty so try/catch
    # is a good idea
    try:
        del r.first_name
    except AttributeError:
        pass

    # Recompute and return first name (after clearing cache)
    f = p.first_name

    This is a direct copy-paste of Django's cached property from
    https://github.com/django/django/blob/2456ffa42c33d63b54579eae0f5b9cf2a8cd3714/django/utils/functional.py#L38-50
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, type=None):
        if instance is None:
            return self
        res = instance.__dict__[self.func.__name__] = self.func(instance)
        return res


class cached_container(object):
    """
    Decorator to cache containers in such a way that only copies are returned
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, instance, type=None):
        if instance is None:
            return self

        cached_var_name = "_cached_container_for_" + self.func.__name__
        self._cached_var_name = cached_var_name

        if cached_var_name not in instance.__dict__:
            instance.__dict__[cached_var_name] = self.func(instance)
        try:
            out = instance.__dict__[cached_var_name].copy()
        except AttributeError:
            out = copy(instance.__dict__[cached_var_name])
        return out

    def __delete__(self, obj):
        delattr(obj, self._cached_var_name)


class cached_dataframe(cached_container):  # pragma: no cover
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        warnings.warn("@cached_dataframe is deprecated.  @Use cached_container")


class pickle_cache_state:
    """
    This is a descriptor that stores optional state for pickle cache
    """

    def __init__(self, mode=None):
        self.set_mode(mode)

    def set_mode(self, mode=None):
        allowed_modes = ["active", "ignore", "refresh", "reset", "memory"]
        if mode not in allowed_modes:
            raise ValueError(f"You must set mode to be one of {allowed_modes}")
        self.mode = mode

    def __get__(self, instance, owner):
        pass

    def __set__(self, instance, value):
        raise NotImplementedError("You cannot set this attribute")


class pickle_cache_mixin:
    """
    Inherit from this mixin to gain ability to enable/disable pickle_cache
    Has the following methods:
    .enable_pickle_cache              # Enables the pickle cache
    .disable_pickle_cache             # Sets pickle cache to reset mode
    .clear_all_default_pickle_cashes  # Removes ALL default-named cache files

    """

    @classmethod
    def enable_pickle_cache(cls):
        for name, obj in vars(cls).items():
            if isinstance(obj, pickle_cache_state):
                obj.set_mode("active")

    @classmethod
    def disable_pickle_cache(cls):
        for name, obj in vars(cls).items():
            if isinstance(obj, pickle_cache_state):
                obj.set_mode("reset")

    @classmethod
    def clear_all_default_pickle_cashes(cls):
        for file in glob.glob("/tmp/*_*-*-*.pickle"):
            os.unlink(file)


class pickle_cached_container:
    """
    This decorator creates cached containers (i.e. lists, dataframes, etc.)
    It will cache at two different levels.  Calling the property
    multiple times on the same object will return a copy
    of the in-memory cached object.
    If a new object is created, accessing the cached attribute
    will first look for a pickle file with the name specified in
    the decorator.  If that file exists, it will be loaded into
    the in-memory cache and returned.  If it does not exist, the
    logic in the decorated method is exectuted with the results
    being saved both in-memory and to pickle.
    Busting the cache is as simple as deleting the attribute.

    Example:

    class Loader:

        ## This is an optionalal class variable you can add
        ## for when you want to temporarily remove caching.
        ## The name of the class attribute does not matter.
        ##
        ## Modes:
        ##   'active': same as not even specifying this class attribute
        ##   'ignore': maintains all pickle files as they were but ignores the cache
        ##   'refresh': Recomputes and refreshes pickle file
        ##   'reset': Alias for refresh
        ##   'memory': Will recompute and cache on object only.  Not to file.
        pcs = ezr.pickle_cache_state(mode='active')

        # Default cache file /tmp/<cls>.<meth>.pickle
        @ezr.pickle_cached_property()
        def df(self):
            # expensive code to create a dataframe or dict or list
            out = my_expensive_function()
            return out
        @ezr.pickle_cached_property('/tmp/account_data.pickle', return_copy=False)
        def my_dict(self):
            '''
            This property will be cached, but any mutation you make to the property
            will actually mutate the cache.  The reason you might want to use this
            is that it saves a possibly expensive memory copy.  This defaults to
            True because mutating a cache can lead to all sorts of weird bugs.
            '''
            # expensive code to create a dataframe or dict or list
            out = my_expensive_function()
            return out
    loader = Loader()
    # Accesses the cached property, computing/storing if necessary
    # Note: will return a copy of the property to avoid mutation.
    df = loader.df
    # Bust the cache for the property.  This will remove the
    # in-memory cache and delete the pickle file.
    del loader.df
    """

    def __init__(self, pickle_file_name=None, return_copy=True):
        """
        This constructs the class that will decorate the property.
        It is used to record state we will need later
        """
        self._pickle_file_name = pickle_file_name
        self.return_copy = return_copy

    def __call__(self, func):
        """
        Once the decorator object has been initiated, this method
        will be called to do the actual decoration.  All it does is
        replace the method definition on the class with an instance
        of this decorator object, which is also a descriptor.  A
        reference to the initial method is stored on the
        decorator/descriptor class.
        """
        self.func = func
        self.cached_var_name = "_pickle_cache_for_" + self.func.__name__
        return self

    @property
    def default_pickle_file_name(self):
        return "/tmp/{}_{}.pickle".format(
            self.func.__qualname__, str(datetime.datetime.now().date())
        )

    @property
    def pickle_file_name(self):
        if self._pickle_file_name:
            return self._pickle_file_name
        else:
            return self.default_pickle_file_name

    def _get_cache_mode(self, instance):
        cache_mode = None
        for att in instance.__class__.__dict__.values():
            # If a cache state attribute was found on the class
            if isinstance(att, pickle_cache_state):
                # Get the mode from the state
                cache_mode = att.mode
                break
        return cache_mode

    def _get_pickle_or_compute(self, instance):
        """
        This method either pulls data from pickle file if it exists
        otherwise it will compute results, populate the pickle file and
        return the computed object.
        """
        # If pickle file exists, load its contents
        if os.path.isfile(self.pickle_file_name):
            with open(self.pickle_file_name, "rb") as buffer:
                obj = pickle.load(buffer)
        # If pickle file doesn't exist, evaluate the wrapped method
        # and save results to pickle file
        else:
            obj = self.func(instance)
            with open(self.pickle_file_name, "wb") as buffer:
                pickle.dump(obj, buffer)

        return obj

    def _get_memory_pickle_or_compute(self, instance):
        """
        Try to get the object trying in this order:
        1) From memory cached on the object
        2) From the pickle file
        3) By running the decorated method

        The retrieved value will be stored on the object if it isn't
        already there
        """
        if self.cached_var_name not in instance.__dict__:
            instance.__dict__[self.cached_var_name] = self._get_pickle_or_compute(
                instance
            )
        return instance.__dict__[self.cached_var_name]

    def _get_memory_or_compute(self, instance):
        """
        Try to get the object trying in this order:
        1) From memory cached on the object
        2) By running the decorated method

        The retrieved value will be stored on the object if it isn't
        already there
        """
        if self.cached_var_name not in instance.__dict__:
            instance.__dict__[self.cached_var_name] = self.func(instance)
        return instance.__dict__[self.cached_var_name]

    def _get_or_compute(self, instance, cache_mode):
        # If ignoring the cache, always call the decorated method
        if cache_mode == "ignore":
            return self.func(instance)
        # If memory, ignore the pickle file but use object caching
        elif cache_mode == "memory":
            return self._get_memory_or_compute(instance)

        # This looks weird, but it is the same thing as deleting
        # the pickle-cached property on the host object.  Doing so
        # will bust the cache. So refreshing busts the cache
        # and repopulates it by computing.
        elif cache_mode in ["refresh", "reset"]:
            self.__delete__(instance)
            return self._get_memory_pickle_or_compute(instance)

        # Otherwise use object and pickle caching
        else:
            return self._get_memory_pickle_or_compute(instance)

    def _copy_object(self, obj):
        try:
            out = obj.copy()
        except AttributeError:
            out = copy(obj)
        return out

    def __get__(self, instance, type=None):
        """
        After decoration the decorated method will be replaced with
        an instance of the decorator/descriptor.  Every time this instance
        is accessed this method will be called to return the value
        of the method, which has been turned into a pickle-backed property.
        """
        # Get the cache mode
        cache_mode = self._get_cache_mode(instance)

        # Grab the object, (persisting to cache if required)
        obj = self._get_or_compute(instance, cache_mode)

        # Copy the object if appropriate
        if self.return_copy:
            return self._copy_object(obj)
        else:
            return obj

    def __delete__(self, instance):
        """
        This method handles busting the cache.
        """
        # Delete the cached copy of the data
        if self.cached_var_name in instance.__dict__:
            del instance.__dict__[self.cached_var_name]

        # Delete the pickle file
        if os.path.isfile(self.pickle_file_name):
            os.unlink(self.pickle_file_name)


class BlobAttr:
    def __init__(self, default, deep=True):
        if deep:
            self.copy_func = deepcopy
        else:
            self.copy_func = copy

        self._default = default
        self.name = None

    @property
    def default(self):
        return self.copy_func(self._default)

    def __get__(self, obj, cls=None):
        if obj is None:
            return
        else:
            return obj._blob_attr_state[self.name]
            # return self.copy_func(obj._blob_attr_state[self.name])

    def __set__(self, obj, value):
        if obj is None:
            return
        else:
            obj._blob_attr_state[self.name] = self.copy_func(value)


class BlobMixin:
    """
    Inherit from this mixin to get serializable attributes.  This mixin
    defines two methods on the inherited class.  .to_blob() and .from_blob()

    These methods will return (deep) copied versions of all BlobAttr instances
    defined on the base class.  All BlobAttr definitions must include a default
    value.
    Look at the output of BlobMixin.example() too see examples
    """

    @staticmethod
    def example():
        return dedent(
            """
            import easier as ezr


            class Parameters(ezr.BlobMixin):
                drums = ezr.BlobAttr({
                    'first': 'Ringo',
                    'last': 'Star',
                })

                bass = ezr.BlobAttr({
                    'first': 'Paul',
                    'last': 'McCartney',
                })


            # Instantiate a default instance and look at parameters
            params = Parameters()
            print(params.drums, params.bass)

            # Change an attribute explicity
            params.drums = {'first': 'Charlie ', 'last': 'Watts'}

            # Update attributes from a blob
            params.from_blob({'bass': {'first': 'Bill', 'last': 'Wyman'}})

            # Dump the updated attributes to a blob
            blob = params.to_blob()
            print(blob)

            # Update the return blob back to defaults
            blob.update(params.blob_defaults)

            # Load the updated blob back into the params
            params.from_blob(blob)

            # Print the updated results
            print(params.drums, params.bass)
        """
        )

    def __init__(self):
        self._blob_attr_state = {}

        for att_name, att_kind in self.__class__.__dict__.items():
            if isinstance(att_kind, BlobAttr):
                att_kind.name = att_name
                self._blob_attr_state[att_name] = att_kind.default

        self._blob_attr_state_defaults = deepcopy(self._blob_attr_state)

    @property
    def blob_defaults(self):
        return deepcopy(self._blob_attr_state_defaults)

    def to_blob(self):
        return {
            name: deepcopy(getattr(self, name)) for name in self._blob_attr_state.keys()
        }

    def from_blob(self, blob, strict=False):
        blob = deepcopy(blob)
        msg = ""
        extra_keys = set(blob.keys()) - set(self._blob_attr_state.keys())
        missing_keys = set(self._blob_attr_state.keys()) - set(blob.keys())
        # TODO: I need to write tests aroud this.  This is new functionality where when not in strict
        # node, extra blob keys just get ignored
        if extra_keys:
            if strict:
                msg += f"\nBad Blob. These keys unrecognized: {list(extra_keys)}"
            else:
                for key in extra_keys:
                    del blob[key]
        if strict and missing_keys:
            msg += f"\nBad Blob.  These required keys not found: {list(missing_keys)}"
        if msg:
            raise ValueError(msg)

        for key, val in blob.items():
            setattr(self, key, val)

        return self


class Scaler(BlobMixin):
    """
    Scales an arry to have values between 0 and 1.
    With min/max appearing at 0/1 respectively.

    Follows the sklearn transformer api.

    The transformer state is (de)serialized with the
    (from/to)_blob methods.
    """

    limits = BlobAttr(None)

    def fit(self, x):
        import numpy as np

        self.limits = [np.min(x), np.max(x)]
        return self

    def _ensure_fitted(self):
        if self.limits is None:
            raise ValueError("You must fit or load params before you can transform")

    def transform(self, x):
        self._ensure_fitted()
        xf = (x - self.limits[0]) / (self.limits[1] - self.limits[0])
        return xf

    def fit_transform(self, x):
        self.fit(x)
        return self.transform(x)

    def inverse_transform(self, x):
        self._ensure_fitted()
        xr = self.limits[0] + x * (self.limits[1] - self.limits[0])
        return xr


def get_logger(name, level="info"):
    import logging
    import daiquiri

    level_map = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }
    allowed_levels = list(level_map.keys())
    if level not in allowed_levels:
        raise ValueError(f"level must be in {allowed_levels}")

    daiquiri.setup(level=level_map[level])
    logger = daiquiri.getLogger(name)
    return logger


def _generate_html_diff_side(words, opcodes, side):
    """Helper function to generate HTML for one side of the diff view."""
    html = []
    for tag, i1, i2, j1, j2 in opcodes:
        if side == "left":
            content = " ".join(words[i1:i2])
            if tag == "equal":
                html.append(f'<span style="color:#333;">{content}</span>')
            elif tag in ("replace", "delete"):
                html.append(
                    f'<span style="background-color:#ffecec; color:#c33;">{content}</span>'
                )
        else:  # right side
            content = " ".join(words[j1:j2])
            if tag == "equal":
                html.append(f'<span style="color:#333;">{content}</span>')
            elif tag in ("replace", "insert"):
                html.append(
                    f'<span style="background-color:#eaffea; color:#282;">{content}</span>'
                )
    return html


def diff_strings(original_text, modified_text, as_html=False, stand_alone=False):
    """
    Compare two text strings and generate a human-readable diff.

    Args:
        original_text: First string to compare
        modified_text: Second string to compare
        as_html: If True, returns HTML for a side-by-side diff view; if False, returns text-based diff
        stand_alone: If True and as_html=True, returns a complete HTML document with proper headers and styling

    Returns:
        A string containing either:
        - Text-based diff with insertions and deletions marked (if as_html=False)
        - HTML for a side-by-side diff view (if as_html=True, stand_alone=False)
        - Complete HTML document with the diff view (if as_html=True, stand_alone=True)

    Examples:
        # Get a text-based diff
        diff = diff_strings("Hello world", "Hello there world")

        # Get HTML for embedding in a notebook or web page
        html_diff = diff_strings("Hello world", "Hello there world", as_html=True)

        # Get a complete HTML document that can be saved to a file
        standalone_html = diff_strings("Hello world", "Hello there world",
                                     as_html=True, stand_alone=True)
        with open('diff.html', 'w') as f:
            f.write(standalone_html)
    """
    import difflib
    import html

    # Split the text into words
    words1 = original_text.split()
    words2 = modified_text.split()

    # Escape HTML characters if needed
    if as_html:
        words1 = [html.escape(word) for word in words1]
        words2 = [html.escape(word) for word in words2]

    # Create a SequenceMatcher object
    matcher = difflib.SequenceMatcher(None, words1, words2)

    # Get the opcodes (operations needed to transform original_text into modified_text)
    opcodes = matcher.get_opcodes()

    if as_html:
        # Generate a side-by-side HTML diff view
        diff_content = ['<div style="display:flex; width:100%;">']

        # Left side (original text)
        diff_content.extend(
            [
                '<div style="flex:1; padding-right:10px;">',
                "<h3>Original</h3>",
                '<pre style="background-color:#f8f8f8; padding:10px; border-radius:5px;">',
            ]
        )
        diff_content.extend(_generate_html_diff_side(words1, opcodes, "left"))
        diff_content.extend(["</pre>", "</div>"])

        # Right side (modified text)
        diff_content.extend(
            [
                '<div style="flex:1; padding-left:10px;">',
                "<h3>Modified</h3>",
                '<pre style="background-color:#f8f8f8; padding:10px; border-radius:5px;">',
            ]
        )
        diff_content.extend(_generate_html_diff_side(words2, opcodes, "right"))
        diff_content.extend(["</pre>", "</div>", "</div>"])

        diff_html = "\n".join(diff_content)

        if stand_alone:
            # Create a complete HTML document
            html_doc = [
                "<!DOCTYPE html>",
                "<html>",
                "<head>",
                "    <meta charset='utf-8'>",
                "    <meta name='viewport' content='width=device-width, initial-scale=1'>",
                "    <title>Text Difference</title>",
                "    <style>",
                "        body { font-family: Arial, sans-serif; padding: 20px; line-height: 1.4; }",
                "        h3 { margin-top: 0; }",
                "        pre { white-space: pre-wrap; word-wrap: break-word; margin: 0; }",
                "    </style>",
                "</head>",
                "<body>",
                diff_html,
                "</body>",
                "</html>",
            ]
            return "\n".join(html_doc)
        else:
            return diff_html
    else:
        # Process the opcodes to create a human-readable diff (original behavior)
        result = []
        for tag, i1, i2, j1, j2 in opcodes:
            if tag == "equal":
                result.append(" ".join(words1[i1:i2]))
            elif tag == "replace":
                result.append(
                    f"[-{' '.join(words1[i1:i2])}] [+{' '.join(words2[j1:j2])}]"
                )
            elif tag == "delete":
                result.append(f"[-{' '.join(words1[i1:i2])}]")
            elif tag == "insert":
                result.append(f"[+{' '.join(words2[j1:j2])}]")

        return " ".join(result)



import textwrap


class Tracker:
    example = textwrap.dedent(
        """
    # REQUIRES CODE IN TWO DIFFERENT NOTEBOOK CELLS
    

    ***
    Note: There can be some weird holovies syncing issues
    between python and javascript for this.  So occational
    page refreshes might be needed.
    
    **** FOR THIS REASON YOU MAY WANT TO USE ezr.StreamPlotter instead ****
    ***

    The way this works is to start one notebook cell
    with the tracker code.  This will remain live
    and will get updated whenever you run the logging cell.

    After the tracking cell has been set up, you can
    run the logging cell multiple times and the
    tracking cell's visualization will be updated.

    # -- Single-trace use case -----------------------

    # In tracking notebook cell
    # (Note: logy can be a bit tricky.  You might need to first run without it.)
    tracker = ezr.Tracker(logy=True)
    tracker.init()

    # In logging cell

    for nn in np.linspace(1, 20, 50):
        x = np.linspace(.1, nn, 500)
        y = 1 / x
        tracker.update(x, y)
        time.sleep(.02)



    # -- Multi-trace use case -----------------------
    # In tracking notebook cell
    ylim = (-2, 2)
    tracker1 = ezr.Tracker(label='+s', ylim=ylim)
    tracker2 = ezr.Tracker(label='-s', ylim=ylim)
    tracker3 = ezr.Tracker(label='+c', ylim=ylim)
    tracker4 = ezr.Tracker(label='-c', ylim=ylim)

    hv.Overlay([
        tracker1.init(),
        tracker2.init(),
        tracker3.init(),
        tracker4.init()
    ]).collate()

    # In logging cell
    for nn in np.linspace(0, 20, 50):
        x = np.linspace(0, nn, 500)
        y = np.sin(2 * np.pi * x / (4))
        y2 = np.cos(2 * np.pi * x / (4))
        tracker1.update(x, y)
        tracker2.update(x, -y)
        tracker3.update(x, y2)
        tracker4.update(x, -y2)
        time.sleep(.01)
    """
    )

    def __init__(self, label="metric", ylim=None, logy=False, width=800, height=400):
        from holoviews.streams import Pipe
        import holoviews as hv

        self.label = label
        self.ylim = ylim

        self.logy = logy
        self.pipe = Pipe(data=None)
        if hv.Store.current_backend == "bokeh":
            self.dmap = hv.DynamicMap(self._plotter, streams=[self.pipe]).opts(
                framewise=True, width=width, height=height, logy=self.logy
            )
        else:
            self.dmap = hv.DynamicMap(self._plotter, streams=[self.pipe]).opts(
                framewise=True, fig_inches=12, aspect=2, logy=self.logy
            )

    def _plotter(self, data):
        import holoviews as hv

        default_val = hv.Curve(([], []))
        if data is None:
            return default_val
        (x, y) = data

        c = hv.Curve((x, y), label=self.label)
        return c

    def init(self):
        return self.dmap

    def update(self, x, y):
        self.pipe.send((x, y))



import os
from typing import Union, Iterable
import re


def heatmap(df, axis=None, cmap="magma", format="{:.1f}"):
    # This will crash if not run in a jupyter notebook.  That's ok.
    display(df.style.background_gradient(axis=axis, cmap=cmap).format(format))


def column_level_flattener(df, level=1, kill_index_names=False):
    """
    Takes a multi-level column dataframe and returns a flattened version.
    Default is to use level=1, but you can use other levels as well.
    Args:
        level: The level of the index you want to use (defaults to 1)
               "smash" will join column levels with an underscore
        kill_index_names: If True, the column/index names will be set to None

    """
    df = df.copy()
    if level == "smash":
        df.columns = ["_".join([str(v) for v in t]) for t in df.columns]
    else:
        df.columns = df.columns.get_level_values(level)
    if kill_index_names:
        df.columns.name = None
        df.index.name = None

    return df


def slugify(
    vals: Union[str, Iterable[str]],
    sep: str = "_",
    kill_camel: bool = False,
    as_dict: bool = False,
):
    """
    Creates slugs out of string inputs.
    """
    if isinstance(vals, str):
        str_input = True
        vals = [vals]
    else:
        str_input = False

    in_vals = list(vals)
    if kill_camel:
        vals = [re.sub(r"([0-9]|[a-z]|_)([A-Z])", r"\1_\2", v) for v in vals]

    out = [re.sub(r"[^A-Za-z0-9]+", sep, v.strip()).lower() for v in vals]
    out = [re.sub(r"_{2:}", sep, v) for v in out]
    out = [re.sub(r"^_", "", v) for v in out]
    out = [re.sub(r"_$", "", v) for v in out]

    if as_dict:
        return dict(zip(in_vals, out))

    if str_input:
        return out[0]
    else:
        return out


def _pandas_time_integer_converter(series_converter, type_str, df_or_ser, columns=None):
    import pandas as pd

    # You don't want to mutate input object
    df_or_ser = df_or_ser.copy()

    if isinstance(columns, str):
        raise ValueError("You must supply a list of columns")

    # The logic to convert each desired timestamp column of a dataframe
    if isinstance(df_or_ser, pd.DataFrame):
        if columns is None:
            # time_cols = [c for (c, v) in df_or_ser.dtypes.items() if v.name == 'datetime64[ns]']
            time_cols = [c for (c, v) in df_or_ser.dtypes.items() if v.name == type_str]
        else:
            time_cols = columns
        for col in time_cols:
            df_or_ser.loc[:, col] = series_converter(df_or_ser.loc[:, col])
        return df_or_ser

    # The logic to convert a series
    elif isinstance(df_or_ser, pd.Series):
        return series_converter(df_or_ser)

    else:
        raise ValueError(
            "You can only pass dataframes or series objects to this function"
        )


def pandas_time_to_utc_seconds(df_or_ser, columns=None):
    """
    Convert pandas Timestamp records to integer unix seconds
    since epoch.

    If columns is specified, only those columns get converted.
    Otherwise, all timestamp columns get converted
    """

    # A function for converting a series
    def series_converter(ser):
        return ser.view("int64") // 10**9

    return _pandas_time_integer_converter(
        series_converter, "datetime64[ns]", df_or_ser, columns
    )


def pandas_utc_seconds_to_time(df_or_ser, columns=None):
    """
    Convert pandas integer fields to timestamps.  The integers must
    contain the number of seconds since unix epoch.

    You must specify the columns you want to transform
    """

    # A function to convert a series
    def series_converter(ser):
        return (ser.astype("int64") * 10**9).view("datetime64[ns]")

    return _pandas_time_integer_converter(series_converter, "int64", df_or_ser, columns)


def localize_utc_to_timezone(time_ser, timezone, return_naive=True):
    """
    Takes a series of utc timestamps and converts them
    to (naive) timestamps of the specified timezone
    """
    time_ser = time_ser.dt.tz_convert(timezone)
    if return_naive:
        time_ser = time_ser.dt.tz_localize(None)
    return time_ser


def events_from_starting_ending(
    *,
    df,
    start_time_col,
    end_time_col,
    delta_cols=None,
    non_delta_cols=None,
    new_time_col_name="time",
    non_numerics_are_index=True,
):
    """
    Converts a dataframe with start and end times into a dataframe of events.
    Numeric delta columns are assumed to be deltas, and are added to the start time
    and substracted at end time.  Non-delta columns are included in the events
    unaltered and can optionally be used as the index.

    Args:
        df: The dataframe to convert

        start_time_col: The name of the column containing the start time

        end_time_col: The name of the column containing the end time

        delta_cols: The names of the columns to be treated as deltas

        non_delta_cols: The names of the columns to be treated as non-deltas

        new_time_col_name: The name of the new event time column

        non_numerics_are_index: If True, the non-numerics columns are combined
                                with the time column in a groupby statement and
                                the deltas are summed by those groups.
    """
    import pandas as pd

    # Turn defaults args into lists
    if delta_cols is None:
        delta_cols = []

    if non_delta_cols is None:
        non_delta_cols = []

    # If no delta columns are specified, assume all non time columns are deltas
    if delta_cols is None:
        cols_to_keep = set(df.columns) - {start_time_col, end_time_col}
        delta_cols = [c for c in df.columns if c in cols_to_keep]

    # Sainity check columns specifications
    if (set(delta_cols).union(non_delta_cols)).intersection(
        {start_time_col, end_time_col}
    ):
        raise ValueError("(Non)delta cols cant contain start or end time col")

    if set(delta_cols).intersection(non_delta_cols):
        raise ValueError("Same column(s) found in both delta and non_delta columns")

    # Create a starting frame with the start times as event times
    df_start = df[[start_time_col] + non_delta_cols + delta_cols].rename(
        columns={start_time_col: new_time_col_name}
    )

    # Create an ending frame with the end times as event times
    df_end = df[[end_time_col] + non_delta_cols + delta_cols].rename(
        columns={end_time_col: new_time_col_name}
    )

    # Negate the delta columns in the ending frame
    df_end[delta_cols] = -df_end[delta_cols]

    # Create the event frame ordered by event time
    df = pd.concat([df_start, df_end], ignore_index=True, sort=False)
    df = df.sort_values(by=new_time_col_name)

    # Set the index if requested
    if non_numerics_are_index:
        index_cols = non_delta_cols + [new_time_col_name]
        df = df.groupby(by=index_cols).sum().sort_index()

    return df


def weekday_string(ser, kind="slug"):
    """
    Transform a pandas series of datetims into strings
    corresponding to their weekday.
    Args:
        ser: a pandas series object of timestamps
        kind: ['tag', 'slug', 'name']
              tag -> ['0_mon', '1_tue', ...]
              slug -> ['Mon', 'Tue', ...]
              name -> ['Monday', 'Tuesday', ...]
    """
    import pandas as pd
    import calendar

    if not isinstance(ser, pd.Series):
        raise ValueError("Input must be a pandas series")

    ser = ser.dt.weekday

    allowed_kinds = ["slug", "name", "tag"]
    if kind not in allowed_kinds:
        raise ValueError(f"kind must be on of {allowed_kinds}")
    if kind == "tag":
        out = [f"{d}_{calendar.day_abbr[d].lower()}" for d in ser]
    elif kind == "slug":
        out = [calendar.day_abbr[d] for d in ser]
    elif kind == "name":
        out = [calendar.day_name[d] for d in ser]

    out = pd.Series(out, index=ser.index)
    return out


def month_string(ser, kind="slug"):
    """
    Transform a pandas series of datetims into strings
    corresponding to their months.
    Args:
        ser: a pandas series object of timestamps
        kind: ['tag', 'slug', 'name']
              tag -> [a_jan', 'b_feb', ...]
              slug -> ['Jan', 'Feb', ...]
              name -> ['January', 'February', ...]
    """
    import calendar
    from string import ascii_lowercase
    import pandas as pd

    if not isinstance(ser, pd.Series):
        raise ValueError("Input must be a pandas series")

    ser = ser.dt.month

    allowed_kinds = ["slug", "name", "tag"]
    if kind not in allowed_kinds:
        raise ValueError(f"kind must be on of {allowed_kinds}")
    if kind == "tag":
        out = [f"{ascii_lowercase[d]}_{calendar.month_abbr[d].lower()}" for d in ser]
    elif kind == "slug":
        out = [calendar.month_abbr[d] for d in ser]
    elif kind == "name":
        out = [calendar.month_name[d] for d in ser]

    out = pd.Series(out, index=ser.index)
    return out


def get_quick_schema_class():
    import pandera as pa

    class QuickSchema:
        dt = pa.dtypes

        def __init__(self, columns, **kwargs):
            """
            A thin wrapper around Pandera DataframeSchema. The main
            use case is to allow more terse schema specification than
            the provided by the Pandera API.

            https://pandera.readthedocs.io/en/stable/reference/generated/pandera.api.pandas.container.DataFrameSchema.html

            The main shortcuts this class provides are that the output
            columns will exactly match the keys of the schema in the
            supplied order.  Additionally, columns can be specified with
            just their types instead of the more verbose pa.Column(...)
            syntax.

            There is a class attribute named dt that is a link to
            pandera datatypes.  This allows for easy tab-completion of types.

            All columns specified as types will result in nullable columns
            that have been coerced to the requested data type.  If you don't
            like this default behavior, pass a pa.Column() object that
            explicitely communicates what you want.

            Once constructed, a QuickSchema object will have a .schema
            attribute that contains the underlying pandera DataFrameSchema.
            This allows you the full power of pandera's api on that schema.

            Apply the schema to a dataframe using the .apply() method, which
            returns a new dataframe

            Args:
                columns: A dict mapping a string columns name into either a
                        full on pa.Columns() instance, or any object that
                        pandera knows how to interpret as a type
                **kwargs: Passed directly to DataframeSchema constructor
            """

            # Save off the column specifications to enable proper column ordering later
            self.col_spec = columns.copy()

            # Apply default ordering
            if "ordered" not in kwargs:
                kwargs["ordered"] = True
            self.ordered = kwargs["ordered"]

            # Apply default coercion
            if "coerce" not in kwargs:
                kwargs["coerce"] = True

            # Initialize a columns dict that Panera will understand
            typed_cols = {}

            # Populate a pandera friendly columns dict
            for key, val in columns.items():
                # If the column type is already a pa.Column type, just use it
                if isinstance(val, type(pa.Column(self.dt.Int16))):
                    typed_cols[key] = val
                # Otherwise, assume it is a type and coerce it to a nullable column
                else:
                    typed_cols[key] = pa.Column(val, nullable=True)

            # Create the Pandera DataFrameSchema object and assign it to the schema attribute
            self.schema = pa.DataFrameSchema(typed_cols, **kwargs)

        @classmethod
        def infer_from_dataframe(cls, df):
            """
            Class method that creates a Quickschema object from a dataframe
            """
            pandera_schema = pa.infer_schema(df)
            spec = {}
            for name, col in pandera_schema.columns.items():
                spec[name] = col.dtype
            return cls(spec)

        @property
        def dtypes(self):
            return dict(self.schema.columns)

        @property
        def ibis_schema(self):
            from ibis.expr import schema as sch

            return sch.from_mapping(
                {
                    col: self.schema.dtypes[col].type
                    for col in self.schema.columns.keys()
                }
            )

        def __repr__(self):
            s = repr(self.schema).replace("DataFrameSchema", "QuickSchema")
            return s

        def __str__(self):
            s = str(self.schema).replace("DataFrameSchema", "QuickSchema")
            return s

        def apply(self, df):
            """
            Apply the schema to retrieve a new dataframe
            """
            if self.ordered:
                df = df[list(self.col_spec.keys())]
            df = df.copy()

            return self.schema(df)

    return QuickSchema


def get_pandas_sql_class():
    import duckdb

    class PandasSql:
        def __init__(self, file=":memory:", overwrite=False, **table_mappings):
            """
            If the file is specified and it contains a database with tables,
            you need not register dataframes. However, any dataframes you register
            will overwrite any existing table of that name.

            Uses duckdb sql dialect

            Args:
            file: str = an optional name for a file for the frame records
            overwrite: bool = When true, this will overwrite the existing database
            **table_mappings: dict = {table_name1: df1, table_name2: df2, ...}
            """
            if file == ":memory:":
                self.file = file
            else:
                self.file = os.path.realpath(os.path.expanduser(file))

            self.conn = self._get_db_connection(self.file, overwrite)
            self.register(**table_mappings)

        def _get_db_connection(self, file, overwrite):
            if overwrite and os.path.isfile(file):
                os.unlink(file)
            conn = duckdb.connect(file)
            return conn

        def register(self, **table_mappings):
            """
            Creates tables from dataframes.
            **table_mappings: dict = {{table_name1: df1, table_name2: df2, ...}}
            """
            if len(table_mappings) == 0:
                return self

            for table_name, df in table_mappings.items():
                self.conn.execute(
                    f"drop table if exists {table_name}; create table {table_name} as select * from df"
                )

        def query(self, sql):
            """
            Returns a dataframe from a sql query.
            (Use .execute() for queries that don't return a value)
            Args:
                sql: str = a string containing the query
            Returns: Pandas Dataframe = a dataframe with the results
            """
            return self.conn.query(sql).to_df()

        def execute(self, sql):
            """
            Runs a sql query (e.g. create, add, drop) that doesn't
            return a value.
            (Use .query() for queries that return a value)
            Args:
                sql: str = a string containing the query
            """
            self.conn.execute(sql)

        @property
        def tables(self):
            """
            A property that returns a frame showing all table names
            """
            df = self.conn.query("show tables").to_df()
            return df.sort_values(by="name")

    return PandasSql


def hex_from_dataframe(df):
    """
    Convert a dataframe to a hex string
    """
    import pickle
    import binascii
    import lzma

    serialized_df = pickle.dumps(df)
    compressed_data = lzma.compress(serialized_df, preset=9)
    hex_encoded_string = binascii.hexlify(compressed_data).decode("utf-8")
    return hex_encoded_string


def hex_to_dataframe(hex_encoded_string):
    """
    Convert a hex string to a dataframe
    """
    import pickle
    import binascii
    import lzma

    # Convert hex string back to binary
    compressed_data = binascii.unhexlify(hex_encoded_string)

    # Decompress using lzma
    serialized_df = lzma.decompress(compressed_data)

    # Unpickle to get the original DataFrame
    df = pickle.loads(serialized_df)

    return df



class Gemini:
    COMPUTE_LOCATION = "us-central1"

    def __init__(
        self,
        model="gemini-2.0-flash-001",
        temperature=0.01,
        top_p=0.95,
        max_output_tokens=8192,
        use_staging=True,
    ):
        import os
        from google import genai
        from google.genai import types

        self.model = model
        self.temperature = temperature
        self.top_p = top_p
        self.max_output_tokens = max_output_tokens

        if use_staging:
            project = os.environ["GOOGLE_CLOUD_PROJECT_STAGING"]
        else:
            project = os.environ["GOOGLE_CLOUD_PROJECT"]

        self.client = genai.Client(
            vertexai=True,
            project=project,
            location=self.COMPUTE_LOCATION,
        )

        self.config = types.GenerateContentConfig(
            temperature=self.temperature,
            top_p=self.top_p,
            max_output_tokens=max_output_tokens,
            response_modalities=["TEXT"],
            safety_settings=[
                types.SafetySetting(
                    category="HARM_CATEGORY_HATE_SPEECH", threshold="OFF"
                ),
                types.SafetySetting(
                    category="HARM_CATEGORY_DANGEROUS_CONTENT", threshold="OFF"
                ),
                types.SafetySetting(
                    category="HARM_CATEGORY_SEXUALLY_EXPLICIT", threshold="OFF"
                ),
                types.SafetySetting(
                    category="HARM_CATEGORY_HARASSMENT", threshold="OFF"
                ),
            ],
        )

    def prompt(self, text):
        from google.genai import types

        contents = [types.Content(role="user", parts=[types.Part.from_text(text=text)])]

        chunk_generator = self.client.models.generate_content_stream(
            model=self.model,
            contents=contents,
            config=self.config,
        )
        return ("".join([c.text for c in chunk_generator])).strip()



def ecdf(
    x,
    N=100,
    inverse=False,
    as_percent=False,
    centered=False,
    folded=False,
    plot=False,
    curve_args=None,
    curve_kwargs=None,
):
    """
    Thin wrapper around statsmodels ecdf.
    Arguments:
        x: array of points for which to find ecdf
        N: The number of output points you want in your ecdf
        inverse: Return 1 - ecdf
        centered: Centers the percentiles at 50%
        folded: Folds the centered percentiles to not exceed 50%
                The result makes the y value be the probability that
                an observsation exceeds the x value in that direction.
    Returns:
        x_out: an array of values
        y_out: an array of percentiles
    """
    import numpy as np
    from statsmodels.distributions.empirical_distribution import ECDF

    x_out = np.linspace(min(x), max(x), N)
    y_out = ECDF(x)(x_out)

    if curve_args is None:
        curve_args = tuple()
    if curve_kwargs is None:
        curve_kwargs = {}

    if centered or folded:
        y_out = y_out - 0.5

    if folded:
        y_out = 0.5 - np.abs(y_out)

    if inverse:
        y_out = 1 - y_out

    if as_percent:
        y_out = 100 * y_out

    if plot:
        import holoviews as hv

        return hv.Curve((x_out, y_out), *curve_args, **curve_kwargs)
    else:
        return x_out, y_out



import contextlib
import os
import re
import warnings


@contextlib.contextmanager
def duck_connection(duck_file_name):
    import duckdb

    con = duckdb.connect(duck_file_name, read_only=False)
    try:
        yield con
    finally:
        con.close()


class Table:
    def __init__(self, name):
        self.name = name

    @property
    def table_name(self):
        return re.sub(r"^df_*", "", self.name)

    def __get__(self, obj, cls):
        if cls is None:  # pragma: no cover
            return

        return obj.query(f"SELECT * FROM {self.table_name}")

    def __set__(self, obj, df=None):
        if df is None:
            return

        if obj._read_only:
            raise ValueError("You can't assign dataframes to a read-only duck")

        with duck_connection(obj.file_name) as con:
            if self.table_name in obj.table_names:
                con.execute(f"DROP TABLE IF EXISTS {self.table_name}")

            con.register("df_in", df)
            con.execute(f"CREATE TABLE {self.table_name} AS SELECT * FROM df_in")


class Duck:
    example = '''
    # Use default file_name
    duck = Duck()

    # Attribute names MUST start with df_ in order to get persisted.
    # The database tablename will be everything following df_
    duck.df_data = df

    # Print a list of all table names in the database
    # You can get frames from these tables by accessing an attribute named "df_<table_name>"
    duck.table_names

    # Accessing the attribute will query all rows from the corresponding
    # table name.
    df = duck.df_data

    # Overwrite any existing specified file
    duck = Duck('./my_database.ddb', overwrite=True)
    duck.df_data = df


    # Run a custom sql query against the database
    df = duck.query("""
        SELECT * FROM table1 JOIN table2 ON field1 LIMIT 10;
    """)
    '''

    def __init__(self, file_name="./duck.ddb", overwrite=False, read_only=False):
        warnings.warn(
            "You are using a deprecated version of Duck.  Use DuckModel instead"
        )
        self.file_name = file_name
        self._read_only = read_only

        if overwrite and os.path.isfile(self.file_name):
            os.unlink(self.file_name)

        if os.path.isfile(self.file_name):
            for table in self.table_names:
                setattr(self, f"df_{table}", None)

    @property
    def table_names(self):
        df = self.query("PRAGMA show_tables")
        return list(df.name)

    def __setattr__(self, name, value):
        import pandas as pd

        if (isinstance(value, pd.DataFrame) or value is None) and name.startswith(
            "df_"
        ):
            setattr(self.__class__, name, Table(name))
        super().__setattr__(name, value)

    def query(self, sql, fetch=True):
        with duck_connection(self.file_name) as con:
            con.execute(sql)
            if fetch:
                df = con.fetchdf()
            else:
                df = None
        return df

    def export_db(self, directory):
        # TODO: Duck db hardcodes the export path.  Try to find a way to make this portable
        directory = os.path.realpath(os.path.expanduser(directory))
        if os.path.isfile(directory) or os.path.isdir(directory):
            raise ValueError(
                f"\n\n {directory!r} already exists.  Cannot overwrite. Nothing done."
            )

        self.query(f"EXPORT DATABASE {directory!r};", fetch=False)

    def import_db(self, directory):
        directory = os.path.realpath(os.path.expanduser(directory))
        if not os.path.isdir(directory):
            raise ValueError(f"\n\n {directory!r} Not found.  Nothing done")

        self.query(f"import DATABASE {directory!r};", fetch=False)



__all__ = ["hv_to_html", "shade", "hist", "cc"]

from typing import Iterable

ALLOWED_REDUCTIONS = {
    "any",
    "count",
    "max",
    "mean",
    "min",
    "std",
    "sum",
}

# loads a holoviews color cycler as cc defaulting to None if not available
# See Defaults section of http://holoviews.org/user_guide/Styling_Plots.html for all available colors
try:

    def get_cc():
        from string import ascii_lowercase
        import holoviews

        cc = type(
            "color",
            (),
            dict(
                zip(ascii_lowercase, holoviews.Cycle().default_cycles["default_colors"])
            ),
        )
        return cc

    cc = get_cc()
except:  # noqa
    cc = None


def hv_to_html(obj, file_name):
    """
    Save a holovies object to html
    :param obj: The holovies object
    :param file_name:  the file name to save. .html will get appended to name
    :return:
    """
    import holoviews as hv

    renderer = hv.renderer("bokeh")

    # Using renderer save
    renderer.save(obj, file_name)


def shade(hv_obj, reduction="any", color=None, spread=False):
    """
    Apply datashading to a holoviews object.

    hv_obj: a holovies object like Curve, Scatter, etc.
    reduction: Most common will be 'any' and 'count'.
               Supply any name here to see list of valid reductions
    color: Mostly used for 'any' aggregation to specify a color
    spread: Smear out points slightly bigger than 1 pixel for easier
            visibility
    """
    import datashader as ds
    from holoviews.operation.datashader import datashade, dynspread

    if reduction not in ALLOWED_REDUCTIONS:
        raise ValueError("Allowed reductions are {}".format(ALLOWED_REDUCTIONS))

    reducer = getattr(ds.reductions, reduction)

    kwargs = dict(aggregator=reducer())
    if color is None and reduction == "any":
        kwargs.update(cmap=["blue"])
    else:
        kwargs.update(cmap=[color])

    obj = datashade(hv_obj, **kwargs)
    if spread:
        obj = dynspread(obj)
    return obj


def hist(x, logx=False, logy=False, label=None, color=None, **kwargs):
    """
    Creates holoviews histrogram object.
        logx=True: Create log spaced bins between min/max
        logy=True: Compute histogram of (1 + x) with y on db scale
        **kwargs are passed to numpy.histogram

    """
    # If logx was specified, create log spaced bins
    import numpy as np
    import holoviews as hv

    if logx:
        nbins = kwargs.get("bins", 10)
        if not isinstance(nbins, int):
            raise ValueError("Bins must be an integer when logx=True")

        range_vals = kwargs.get("range", None)
        if range_vals:
            minval, maxval = range_vals
        else:
            minval, maxval = np.min(x), np.max(x)

        bins = np.logspace(np.log10(minval), np.log10(maxval), nbins)
        kwargs.update(bins=bins)

    # Build up kwargs for the holoviews call
    hv_kwargs = {}
    if label is not None:
        hv_kwargs["label"] = label

    # If logy was specified, create a histogram of the db of (counts + 1)
    if logy:
        counts, edges = np.histogram(x, **kwargs)
        counts = 10 * np.log10(1 + counts)
        c = hv.Histogram((counts, edges), vdims="dB of (counts + 1)", **hv_kwargs)
    # If not logy, just to a histogram of counts
    else:
        c = hv.Histogram(np.histogram(x, **kwargs), **hv_kwargs)

    # Default the x axis to log if logx was specified
    if logx:
        c = c.options(logx=True)
    c = c.options(alpha=0.3)
    if color is not None:
        c = c.options(color=color)
    return c


def beta_plots(
    wins: Iterable,
    losses: Iterable,
    labels: Iterable,
    legend_position="right",
    alpha=0.5,
    normed=False,
    xlabel=None,
    ylabel=None,
):
    """
    Make beta plots for win/loss type scenarios.  The wins/losses are provided in arrays.
    Each element of the array corresponds to a specific win/loss scenario you want plotted.

    So, to just determine the beta distirbution of a single scenario, these should be
    one-element arrays.

    Args:
        wins: the number of "wins"
        losses: the number of "losses"
        labels: The label for each scenario
        legend_postion: Where to put the legend
        alpha: the opacity of the area plots
        xlabel: The x label [default "Win Perdentage"]
        ylabel: The y label [default, "Density"]
    """
    from scipy.stats import beta
    import numpy as np
    import holoviews as hv

    if xlabel is None:
        xlabel = "Win Percentage"

    if ylabel is None:
        ylabel = "Density"

    c_list = []
    x = np.linspace(0, 1, 500)
    xpoints = []
    ypoints = []
    for won, lost, label in zip(wins, losses, labels):
        dist = beta(won + 1, lost + 1)
        y = dist.pdf(x)
        if normed:
            y_max = np.max(y)
        else:
            y_max = 1.0
        y = y / y_max
        win_frac = won / (won + lost + 1e-12)
        xpoints.append(win_frac * 100)
        ypoints.append(dist.pdf(win_frac) / y_max)
        c = hv.Area((100 * x, y), xlabel, ylabel, label=label).options(alpha=alpha)
        c_list.append(c)

    c1 = hv.Overlay(c_list).options(legend_position="right")
    c2 = hv.Scatter((xpoints, ypoints), xlabel, ylabel).options(
        color="black", size=8, tools=["hover"]
    )
    return (c1 * c2).options(legend_position=legend_position)


class Animator:
    """
    Creates animated plots with holoviews.

    Constructor Args:
        dynamic_range: bool = if set to True, auto-scale the chart each time it is drawn

    # ------------ Simple Example -------------------
    # Define a function to make your plots.
    # Data will be whatever object is passed to animator.send()
    def myplot(data):
        return hv.Curve(data)

    # Pass your plotting function to the animator constructor
    animator = Animator(myplot)

    # Simply send the animator updated versions of your data
    # to update the plot
    t0 = np.linspace(0, 6., 100)
    for delta in np.linspace(0, 3, 30):
        t = t0 - delta
        y = 2 + np.sin(t)
        animator.send((t, y))

    # ------------ Advanced Example -------------------
    def myplot(data):
        # First two elements of data are to be ploted
        # Third element of data is going to update a label
        c = hv.Curve(data[:2], 'a', 'b', label=f'hello {data[-1]}').options(color=ezr.cc.c, logy=True)
        c *= hv.Scatter((data[0], 2 * data[1])).options(color=ezr.cc.d)
        return c


    # Send data for animation
    for ind, delta in enumerate(np.linspace(0, 3, 300)):
        t = t0 - delta
        y = 2 + np.sin(t)
        # Can control when animations are drawn with this
        if ind % 1 == 0:
            animator.send((t, y))
    """

    def __init__(self, plot_func, dynamic_range=True, plot_every=1):
        import holoviews as hv
        from holoviews.streams import Pipe

        # Store the user-defined plot function
        self.plot_func = plot_func
        self.dynamic_range = dynamic_range
        self.pipe = Pipe(data=[])
        self.data_has_been_sent = False
        self.dmap = hv.DynamicMap(self.plot_wrapper, streams=[self.pipe])
        self.plot_every = plot_every
        self.plot_count = 0

    def plot_wrapper(self, *args, **kwargs):
        data = kwargs.get("data", ([0], [0]))
        hv_obj = self.plot_func(data)
        if self.dynamic_range:
            hv_obj = hv_obj.opts(norm=dict(framewise=True))
        return hv_obj

    def send(self, data):
        from IPython.display import display

        if self.plot_count % self.plot_every == 0:
            self.pipe.send(data)
            if not self.data_has_been_sent:
                display(self.dmap)
                self.data_has_been_sent = True
        self.plot_count += 1



from collections import Counter
from contextlib import contextmanager
import datetime


class GlobalClock:
    delta = Counter()
    active_start_times = dict()

    def __init__(self):
        pass

    @contextmanager
    def running(self, *names):
        """
        Args:
            *names: *args of string names for clocks that should be running in the context
        """
        self.start(*names)
        yield
        self.stop(*names)

    @contextmanager
    def paused(self, *names):
        """
        Args:
            *names: *args of string names for clocks that should be paused in the context
        """
        allowed_names = set(self.active_start_times.keys())
        bad_names = set(names) - allowed_names
        if bad_names:
            raise ValueError(
                f"Trying to pause clocks {list(bad_names)} that aren't running."
            )
        if not names:
            raise ValueError("You must specify at least one clock name to pause")
        self.stop(*names)
        yield
        self.start(*names)

    def start(self, *names):
        """
        Args:
            *names: *args of string names for clocks that should be started
        """
        if not names:
            raise ValueError("You must provide at least one name to start")

        for name in names:
            if name not in self.active_start_times:
                self.active_start_times[name] = datetime.datetime.now()

    def stop(self, *names):
        """
        Args:
            *names: *args of string names for clocks that should be stopped

            If no names are provided, then stop all active clocks.
        """
        ending = datetime.datetime.now()
        if not names:
            names = list(self.active_start_times.keys())
        for name in names:
            if name in self.active_start_times:
                starting = self.active_start_times.pop(name)
                self.delta.update({name: (ending - starting).total_seconds()})

    def reset(self, *names):
        """
        Args:
            *names: *args of string names for clocks that should be reset

            If no names are provided, then stop and reset all clocks.
        """
        if not names:
            names = list(self.active_start_times.keys())
            names.extend(list(self.delta.keys()))
        for name in names:
            if name in self.delta:
                self.delta.pop(name)
            if name in self.active_start_times:
                self.active_start_times.pop(name)

    def get_time(self, *names):
        """
        Args:
            *names: *args of string names for clocks from which you want the current time
        Returns:
            If one name is provided, a floating point number with the time for that clock.
            If multiple names, a dict of {name: time} is returned for specified names
            If no names are provided, then a dict of all running clocks is returned
        """
        ending = datetime.datetime.now()
        if not names:
            names = list(self.delta.keys())
            names.extend(list(self.active_start_times.keys()))

        delta = Counter()
        for name in names:
            if name in self.delta:
                delta.update({name: self.delta[name]})
            elif name in self.active_start_times:
                delta.update(
                    {name: (ending - self.active_start_times[name]).total_seconds()}
                )
        if len(delta) == 1:
            return delta[list(delta.keys())[0]]
        else:
            return dict(delta)

    def __str__(self):
        """
        Print the current state of the clock.  Only include outputs for clocks that
        have been started and stopped.
        """
        records = sorted(self.delta.items(), key=lambda t: t[1], reverse=True)
        records = [("%0.6f" % r[1], r[0]) for r in records]

        if records:
            out_list = ["{: <15s}{}".format("seconds", "name")]
        else:
            out_list = []

        for rec in records:
            out_list.append("{: <15s}{}".format(*rec))

        return "\n".join(out_list)

    def __repr__(self):
        return self.__str__()


class Clock(GlobalClock):
    def __init__(self):
        # Override class attributes with instance attributes
        self.delta = Counter()
        self.active_start_times = dict()



import contextlib
import datetime
import sys


class OutStream(object):  # pragma no cover
    """
    This class exisist for easing testing of sys.stdout and doesn't
    need to be tested itself
    """

    def write(self, *args, **kwargs):
        sys.stdout.write(*args, **kwargs)


class TimerResult(object):
    def __init__(self, label, starting, ending=None, seconds=None):
        self.label = label
        self.starting = starting
        self.ending = ending
        self.seconds = seconds

    def __str__(self):
        return "__time__,{},{}".format(self.seconds, self.label)

    def __repr__(self):
        return self.__str__()


@contextlib.contextmanager
def Timer(name="", silent=False, pretty=False):
    """
    A context manager for timing sections of code.
    :type name: str
    :param name: The name you want to give the contextified code
    :type silent: bool
    :param silent: Setting this to true will mute all printing
    :type pretty: bool
    :param pretty: When set to true, prints elapsed time in hh:mm:ss.mmmmmm
    Example
    ---------------------------------------------------------------------------
    # Example code for timing different parts of your code
    import time
    from pandashells import Timer
    with Timer('entire script'):
        for nn in range(3):
            with Timer('loop {}'.format(nn + 1)):
                time.sleep(.1 * nn)
    # Will generate the following output on stdout
    #     col1: a string that is easily found with grep
    #     col2: the time in seconds (or in hh:mm:ss if pretty=True)
    #     col3: the value passed to the 'name' argument of Timer

    __time__,2.6e-05,loop 1
    __time__,0.105134,loop 2
    __time__,0.204489,loop 3
    __time__,0.310102,entire script

    ---------------------------------------------------------------------------
    # Example for measuring how a piece of of code scales (measuring "big-O")
    import time
    from pandashells import Timer

    # initialize a list to hold results
    results = []

    # run a piece of code with different values of the var you want to scale
    for nn in range(3):
        # time each iteration
        with Timer('loop {}'.format(nn + 1), silent=True) as timer:
            time.sleep(.1 * nn)
        # add results
        results.append((nn, timer))

    # print csv compatible text for further pandashells processing/plotting
    print 'nn,seconds'
    for nn, timer in results:
        print '{},{}'.format(nn,timer.seconds)
    """
    stream = OutStream()
    result = TimerResult(name, starting=datetime.datetime.now())
    yield result
    result.ending = datetime.datetime.now()
    dt = result.ending - result.starting
    result.seconds = dt.total_seconds()
    dt = dt if pretty else result.seconds
    if not silent:
        stream.write("__time__,{},".format(dt))
        if name:
            stream.write("%s\n" % name)



import abc
import os
import easier as ezr
from typing import Optional

# from google.cloud import bigquery
# from google.oauth2 import service_account
import pandas_gbq
import pandas as pd
import numpy as np

example = """
from easier.bigquery import BQAppender, BQTable, BQDataset
import pandas as pd
import datetime


class MyAppender(BQAppender):
    def get_dataframe(self, **kwargs):
        # self.validate_kwargs(**kwargs)

        df = pd.DataFrame(
            {
                'my_str': ['a', 'b'],
                'my_int': [1, 2],
                'my_float': [3., 4.],
            }
        )
        time = kwargs['time']
        df['time'] = time
        return df

    def validate_kwargs(self, **kwargs):
        if 'time' not in kwargs:
            raise ValueError('kwargs must contain "time"')


class MyDataset(BQDataset):
    dataset_id = 'rob'

    my_table = BQTable(
        name='my_table',
        appender_class=MyAppender,
        schema={
            'time': 'TIMESTAMP',
            'my_str': 'STRING',
            'my_int': 'INTEGER',
            'my_float': 'FLOAT',
        }

    )

    def push(self):
        time = datetime.datetime.now()
        for table in self.tables:
            table.append(time=time)


project_id = 'ambition-analytics'
ds = MyDataset(project_id, create_missing_dataset=True)
ds.push()
"""


class BQAppender(metaclass=abc.ABCMeta):
    """
    A BQTable instance needs to know how to grab data to append to the database.
    One of the arguments it takes is a appender class.  This is the base class you should
    derive from when creating your own Appenders.
    """

    @abc.abstractmethod
    def get_dataframe(self, **kwargs):
        """
        This method must be implemented.  It should contain the implementation of how
        you want to grab the data you want to push into the database.

        Any kwargs can be supplied to this method.  If you wish to validate the kwargs to this method,
        you can do so using the validate_kwargs(*kwargs) method.
        """

    def validate_kwargs(self, **kwargs):
        """
        Override this method to provide (optional) validation on the kwargs sent to the finalize_frame()
        method.
        """


class BQTable:
    """
    BQTable is a descriptor you should use when defining your BQDataset class.  It provides an
    interface for appending and accessing data from a particular table in your Bigquery Dataset.
    """

    # By default we use "DAY" parititoning for bigquery tables.
    _PARTITIONING_TYPE = "DAY"

    # This provides a lookup between pandas dtypes and bigquery database types.
    # It is used to validate schemas before appendin gdata
    _TYPE_LOOKUP = {
        pd.Timestamp: "TIMESTAMP",
        np.int64: "INTEGER",
        np.float64: "FLOAT",
        str: "STRING",
    }

    def __init__(
        self, name: str, schema: dict, appender_class: Optional[BQAppender] = None
    ):
        """
        Args:
            name: the name of the databse table to create/use in Bigquery
            schema: A dictionary with keys of field names and values of field types.
            appender_class: The class to use for appending data.
        """
        self.name = name
        self.schema = schema
        self.existance_confirmed = False
        self.appender_class = appender_class

    def __str__(self):
        return f"BQTable({self.name})"

    def __repr__(self):
        return self.__str__()

    def _table_exists(self, dataset):
        """
        Returns whether or not this table exists
        """
        return self.name in [
            t.table_id for t in dataset.client.list_tables(dataset.dataset_id)
        ]

    def _ensure_exists(self, dataset):
        """
        If it doesn't already exist, create this table on the provided dataset using the defined schema.
        """
        from google.cloud import bigquery

        if not self._table_exists(dataset):
            table_ref = dataset.dataset.table(self.name)
            schema = list(
                bigquery.SchemaField(field, dtype)
                for (field, dtype) in self.schema.items()
            )
            table = bigquery.Table(table_ref, schema=schema)
            table.partitioning_type = self._PARTITIONING_TYPE
            dataset.client.create_table(table)
        self.existance_confirmed = True

    def __get__(self, dataset, dataset_class):
        """
        This is what makes this class a descriptor
        """
        self._dataset = dataset
        self._ensure_exists(dataset)
        return self

    def _check_dataframe(self, df):
        """
        Checks to make sure a dataframe is valid for appending to the bigquery table
        """
        if df.empty:
            return

        expected_cols = set(self.schema.keys())
        received_cols = set(df.columns)
        missing_cols = expected_cols - received_cols
        if missing_cols:
            raise ValueError(f"Dataframe missing the following columns: {missing_cols}")

        bad_list = []
        for col, col_type in self.schema.items():
            mapped_type = self._TYPE_LOOKUP.get(type(df[col].iloc[0]))
            if col_type != mapped_type:
                bad_list.append(col)
        if bad_list:
            raise ValueError(f"Bad columns types for these columns: {bad_list}")

    def append_dataframe(self, df):
        """
        Call this method if you have a dataframe you want to manually append to the table.  Usually
        you won't want to use this method, but instead use the .append() method which will use your
        defined Appender class to create the dataframe and then append it to the bigquery table.
        """
        # Nothing to do if the dataframe is empty
        if df.empty:
            return

        # Make sure the dataframe looks right
        self._check_dataframe(df)

        # Make sure the frame is organized to be consistent with schema definition
        df = df[list(self.schema.keys())]

        # Push the dataframe to bigquery
        pandas_gbq.to_gbq(
            df,
            self.name_in_project,
            project_id=self._dataset.project_id,
            if_exists="append",
            progress_bar=False,
        )

    def get_dataframe(self, **kwargs):
        """
        This will pull in the source dataframe from this tables appender.  This is the dataframe
        that is used as the source for appending data to the bigquery table.
        """
        # Make sure that an appender class has been defined
        if self.appender_class is None:
            raise ValueError(
                "You must provide an appender class to the constructor if you want to call append()"
            )

        # Instantiate the appender class
        appender = self.appender_class()

        # Make sure the kwargs are valid
        appender.validate_kwargs(**kwargs)

        # Grab the dataframe
        df = appender.get_dataframe(**kwargs)

        return df

    def append(self, dry_run=False, **kwargs):
        """
        This is the preferred method for appending data to the bigquery database.  It will
        use the supplied Appender class to load a frame whose data will be pushed to bigquery.
        You can supply optional keyword arguments that will be passed to the .finalize_frame() method
        of your appender class.  These kwargs provide the ability to add whatever final/custom processing
        you want to your dataframe.
        """
        df = self.get_dataframe(**kwargs)

        # Append it to bigquery
        if not dry_run:
            self.append_dataframe(df)

    @ezr.cached_property
    def full_name(self):
        """
        This is the full name of the table.  It is what you should use in the
        FROM clause of any sql query.
        """
        return f"{self._dataset.project_id}.{self.name_in_project}"

    @ezr.cached_property
    def name_in_project(self):
        """
        The name of this table within the project
        """
        return f"{self._dataset.dataset_id}.{self.name}"

    @ezr.cached_property
    def latest_partition_time(self):
        """
        An attribute with a cached latest parition time
        """
        return self.get_latest_partition_time()

    @ezr.cached_container
    def latest_partition_frame(self):
        return self.get_latest_partition_frame()

    def get_latest_partition_frame(self):
        """
        An attribute with a cached dataframe of the latest parition data
        """
        sql = """
            SELECT
                *
            FROM
                {table}
            WHERE
                _PARTITIONDATE = "{partition_date_iso}"
        """.format(
            table=self.full_name,
            partition_date_iso=self.latest_partition_time.date().isoformat(),
        )
        df = self._dataset.query(sql)
        return df

    def get_latest_partition_time(self):
        """
        Grabs the most recent partition time for the table
        """
        sql_template = """
            SELECT
                _PARTITIONDATE AS partition_date
            FROM
                {full_name}
            WHERE
                _PARTITIONDATE >= DATE_SUB(CURRENT_DATE(), INTERVAL {days} DAY)
            ORDER BY
                partition_date DESC
            LIMIT 1
        """
        # This will exponentially increase the lookback time for partitions up to 256 days.
        days = 1
        while days <= 256:
            sql = sql_template.format(full_name=self.full_name, days=days)
            df = self._dataset.query(sql)
            if not df.empty:
                return df.partition_date.iloc[0].to_pydatetime()
            days = 2 * days
        raise RuntimeError(
            "To reduce data-size you can't query partitioned more than 256 days ago"
        )

    def get_latest_data(self, time_field_name):
        """
        If you have timestamped data in the database, this will grab all data with the latest timestamp
        for the latest partition.
        """

        sql = """
            WITH latest_partition AS (
                SELECT
                    *
                FROM
                    {table}
                WHERE
                    _PARTITIONDATE = "{partition_date_iso}"
            )
            SELECT
                *
            FROM
                latest_partition
            WHERE
                {time_field_name} = (SELECT MAX({time_field_name}) FROM latest_partition)
        """.format(
            table=self.full_name,
            time_field_name=time_field_name,
            partition_date_iso=self.latest_partition_time.date().isoformat(),
        )
        df = self._dataset.query(sql)
        return df


class BQDataset:
    def __init__(
        self, project_id, *, dataset_id=None, create_missing_dataset=False, **kwargs
    ):
        """
        A class that knows how to define and work with tables in a BigQuery dataset.

        Args:
            project_id: The project id this dataset belongs to
            dataset_id: The (optional) id for the dataset
            create_missing_dataset: Setting this flag will create non-existing datasets.
            kwargs: Any arguments you pass here will get passed to a .constructor_kwargs attribute.
        """
        # TODO: I don't like the interface I've chosen here.  I don't think a class variable is a good way
        # to go because it inhibits flexibility.  Also the way I chose project_id as an argument
        # and dataset_id as an optional argument is weird.  This needs improvement.

        if dataset_id is not None:
            self.dataset_id = dataset_id

        self.project_id = project_id
        self.constructor_kwargs = kwargs

        if not hasattr(self, "dataset_id"):
            raise ValueError(
                "You must set dataset_id as class-variable "
                "on your derived class or pass data_set_id to the constructor"
            )

        self._ensure_dataset_exists(create_missing_dataset)

    @ezr.cached_property
    def full_dataset_name(self):
        return f"{self.project_id}.{self.dataset_id}"

    @ezr.cached_property
    def all_dataset_names(self):
        return [d.dataset_id for d in self.client.list_datasets()]

    def _ensure_dataset_exists(self, create_missing_dataset):
        if self.dataset_id not in self.all_dataset_names:
            if create_missing_dataset:
                self._create_dataset()
            else:
                msg = (
                    f'Dataset "{self.dataset_id}" does not exist. '
                    "You can set create_missing_dataset=True to create it"
                )
                raise RuntimeError(msg)

    def ensure_all_tables_exist(self):
        for table in self.tables:
            table._ensure_exists(self)

    def _create_dataset(self):
        from google.cloud import bigquery

        dataset = bigquery.Dataset(self.full_dataset_name)
        dataset.location = "US"
        self.client.create_dataset(dataset, timeout=30)

    @property
    def auth_file(self):
        # TODO: Try to figure out how to init the contents of the auth file into env vars
        auth_file = None
        config_dir = os.path.expanduser("~/.config/gbq")
        if os.path.isdir(config_dir):
            auth_file_list = os.listdir(config_dir)
            if config_dir:
                auth_file = os.path.join(config_dir, auth_file_list[0])
        return auth_file

    @classmethod
    def get_table_names(cls):
        """
        A method to return all table names
        """
        table_names = []
        for name, obj in vars(cls).items():
            if isinstance(obj, BQTable):
                table_names.append(name)
        return table_names

    @ezr.cached_property
    def tables(self):
        """
        A list attribute containing all tables
        """
        tables = []
        for name in self.get_table_names():
            tables.append(getattr(self, name))
        return tables

    @property
    def on_dev_machine(self):
        """
        This will probably change when I do this for real, but for now, I check if I'm
        in a cloud function or not by checking for the existence of an auth file
        """
        return self.auth_file is not None

    @ezr.cached_property
    def client(self):
        """
        An attribute holding the google client
        """
        from google.cloud import bigquery
        from google.oauth2 import service_account

        client_kwargs = {"project": self.project_id}
        if self.on_dev_machine:
            client_kwargs.update(
                credentials=service_account.Credentials.from_service_account_file(
                    self.auth_file
                )
            )
        client = bigquery.Client(**client_kwargs)
        return client

    @ezr.cached_property
    def dataset(self):
        """
        An attribute holding the google dataset
        """
        return self.client.get_dataset(self.full_dataset_name)

    def query(self, sql_query, show_progress=False):
        """
        Run BigTable queries on this dataset. Tables can be referenced in FROM and JOIN SQL clauses
        using their .full_name attributes injected into a query template.
        """
        if show_progress:
            progress_bar_type = "tqdm"
        else:
            progress_bar_type = None

        df = pandas_gbq.read_gbq(
            sql_query,
            project_id=self.project_id,
            progress_bar_type=progress_bar_type,
        )
        return df



from collections import namedtuple
from typing import List, Dict, Any, Tuple
import copy
import functools
import importlib
import os
import urllib.parse


def pg_creds_from_env(kind="dict", force_docker=False):
    """
    Pulls postgres credentials from the environment.  If env vars don't exist,
    it will default to the default docker creds.  You can force this behaviour
    by adding force_docker=True
    """
    allowed_kinds = ["url", "dict"]

    if kind not in allowed_kinds:
        raise ValueError(f"Allowed kinds are {allowed_kinds}")

    if force_docker:
        env = {}
    else:
        env = os.environ.copy()

    creds = {
        "host": env.get("PGHOST", "db"),
        "port": env.get("PGPORT", "5432"),
        "database": env.get("PGDATABASE", "postgres"),
        "user": env.get("PGUSER", "postgres"),
        "password": env.get("PGPASSWORD", "postgres"),
    }
    # URL encode username and password to handle special characters
    encoded_user = urllib.parse.quote(creds["user"])
    encoded_password = urllib.parse.quote(creds["password"])
    url = f"postgresql://{encoded_user}:{encoded_password}@{creds['host']}:{creds['port']}/{creds['database']}"

    if kind == "dict":
        return creds
    else:
        return url


class PG:
    _sql = None
    _context = None
    _conn_kwargs = None
    _raw_results = None
    _raw_columns = None

    def __init__(self, **kwargs):
        """
        kwargs = dict(host=None, user=None, password=None, dbname=None)

        Any kwargs that are not supplied will be loaded from the environment.
        Environment variables should be named according to the psql convention:
            kwarg:  environment_var_name
            'host': 'PGHOST',
            'user': 'PGUSER',
            'password': 'PGPASSWORD',
            'dbname': 'PGDATABASE'

        Using with Django
        After django is loaded, simply construct with

            pg = PG(use_django=True)

        Tables can be shown with
            pg.table_names()

        Django queries can be run with one line
        df = PG(use_django=True).query('SELECT * FROM my_table').df
        """
        # See if db info should be loaded from django
        use_django = kwargs.get("use_django", False)

        # If so, replace kwargs with values from django settings
        if use_django:
            self.safe_import("django")
            from django.conf import settings

            db = settings.DATABASES["default"]
            kwargs = {
                "host": db["HOST"],
                "user": db["USER"],
                "password": db["PASSWORD"],
                "dbname": db["NAME"],
            }

        env_translator = {
            "host": "PGHOST",
            "user": "PGUSER",
            "password": "PGPASSWORD",
            "dbname": "PGDATABASE",
        }
        conn_kwargs = {
            key: kwargs.get(key, os.environ.get(env_translator[key], None))
            for key in env_translator.keys()
        }

        bad_keys = []
        for key in conn_kwargs.keys():
            if conn_kwargs[key] is None:
                bad_keys.append(key)
        if bad_keys:
            raise ValueError(
                f"The following connections params not specified {bad_keys}"
            )

        self._conn_kwargs = conn_kwargs

    @classmethod
    def queryset_to_sql(cls, queryset):
        """
        Transform a queryset into pretty sql that can be copy-pasted directly
        into pg-admin
        """
        # Do imports here to avoid dependencies
        sqlparse = cls.safe_import("sqlparse")
        cls.safe_import("django")
        from django.db import connection

        # Compile the query to python db api
        sql, sql_params = queryset.query.get_compiler(using=queryset.db).as_sql()

        # Translate the python query spec into a postgres query
        with connection.cursor() as cur:
            query = cur.mogrify(sql, sql_params)

        # Make the query pretty and return it
        query = sqlparse.format(query, reindent=True, keyword_case="upper")
        return query

    def schema_names(self):
        # Run schema query in a copied version of self so as not to mess with current query
        # on this object
        pg = copy.deepcopy(self)
        return pg.query("SELECT nspname FROM pg_catalog.pg_namespace").to_dataframe()

    @functools.lru_cache()
    def table_names(self, schema_name="public"):
        # Run table query in a copied version of self so as not to mess with current query
        # on this object
        pg = copy.deepcopy(self)
        df = pg.query(
            f"SELECT table_name FROM information_schema.tables WHERE table_schema='{schema_name}'"
        ).to_dataframe()
        df = df.sort_values(by="table_name")
        return df

    def query(self, sql) -> "PG":
        """
        sql: SQL query
        """
        self._sql = sql
        return self

    def run(self) -> "PG":
        """
        Runs the query on the database populating instance variables with results
        """
        psycopg2 = self.safe_import("psycopg2")
        with psycopg2.connect(**self._conn_kwargs) as connection:
            with connection.cursor() as cursor:
                cursor.execute(self._sql)
                try:
                    self._raw_results = list(cursor.fetchall())
                    self._raw_columns = [col[0] for col in cursor.description]
                except psycopg2.ProgrammingError as e:
                    if str(e) == "no results to fetch":
                        self._raw_results = []
                        self._raw_columns = []
                    else:  # pragma: no cover  No expected to hit this, but raise just in case
                        raise
        return self

    def reset(self):
        self._raw_columns = None
        self._raw_results = None

    @property
    def _results(self) -> List[Tuple]:
        """
        A list of raw result tuples
        """
        if self._raw_results is None:
            self.run()
        return self._raw_results

    @property
    def columns(self) -> List[str]:
        if self._raw_columns is None:
            self.run()
        return self._raw_columns

    def as_tuples(self) -> List[Tuple]:
        """
        :return: Results as a list of tuples
        """
        self.reset()
        return self._results

    def as_dicts(self) -> List[Dict[str, Any]]:
        """
        :return: Results as a list of dicts
        """
        self.reset()
        return [dict(zip(self.columns, row)) for row in self._results]

    def as_named_tuples(self, named_tuple_name="Result") -> List[Any]:
        """
        :return: Results as a list of named tuples
        """
        self.reset()
        # Ignore typing in here because of unconventional namedtuple usage
        nt_result = namedtuple(named_tuple_name, self.columns)  # type: ignore
        return [nt_result(*row) for row in self._results]  # type: ignore

    @classmethod
    def safe_import(cls, module_name, package=None):
        try:
            imported = importlib.import_module(module_name, package=package)
        except (
            ImportError
        ):  # pragma: no cover.  Not going to uninstall pandas to test this
            raise ImportError(
                f"\n\nNope! This method requires that {module_name} be installed.  You know what to do."
            )

        return imported

    def as_dataframe(self) -> Any:
        """
        :return: Results as a pandas dataframe
        """
        self.reset()
        pd = self.safe_import("pandas")

        return pd.DataFrame(self._results, columns=self.columns)

    def to_tuples(self) -> List[Tuple]:
        """
        alias
        """
        return self.as_tuples()

    def to_dicts(self) -> List[Dict[str, Any]]:
        """
        alias
        """
        return self.as_dicts()

    def to_named_tuples(self) -> List[Any]:
        """
        alias
        """
        return self.as_named_tuples()

    def to_dataframe(self) -> Any:
        """
        alias
        """
        return self.as_dataframe()

    @property
    def df(self):
        return self.to_dataframe()

    @property
    def sql(self) -> str:
        sqlparse = self.safe_import("sqlparse")
        psycopg2 = self.safe_import("psycopg2")
        with psycopg2.connect(**self._conn_kwargs) as connection:
            with connection.cursor() as cursor:
                sql, params = self._get_prepared_query()
                query = cursor.mogrify(sql, params)
                query = sqlparse.format(query, reindent=True, keyword_case="upper")
        return query


def sql_file_to_df(file_name="sql_query.sql", context_dict=None):
    """
    Load and execute a SQL query from a file, returning the results as a DataFrame.
    Always connects to the postgres database with credentials from the environment.
    Parameters:
    -----------
    file_name : str
        Path to the SQL file (default: 'sql_query.sql')
    context_dict : dict, optional
        Dictionary of variables to use for template rendering.

    Returns:
    --------
    pandas.DataFrame
        Results of the SQL query
    """
    from pathlib import Path
    import jinja2
    import easier as ezr

    if context_dict is None:
        context_dict = {}

    # Expand user directory and get absolute path
    file_path = Path(file_name).expanduser().resolve()

    # Extract the directory the file lives in
    dir_path = file_path.parent

    jinja_env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(dir_path.as_posix()),
        autoescape=jinja2.select_autoescape(),
    )
    template = jinja_env.get_template(file_path.name)
    text = template.render(**context_dict)

    pg = ezr.PG()
    pg.query(text)
    df = pg.to_dataframe()
    return df


def sql_string_to_df(query, context_dict=None):
    """
    Execute a SQL query string with Jinja templating, returning the results as a DataFrame.
    Alwyas connects to the postgres database with credentials from the environment.

    Parameters:
    -----------
    query : str
        SQL query string with optional Jinja template variables
    context_dict : dict, optional
        Dictionary of variables to use for template rendering.
        Defaults to globals().

    Returns:
    --------
    pandas.DataFrame
        Results of the SQL query
    """
    import jinja2
    import easier as ezr

    if context_dict is None:
        context_dict = {}

    # Create a Jinja environment with a string loader
    jinja_env = jinja2.Environment(
        loader=jinja2.BaseLoader(), autoescape=jinja2.select_autoescape()
    )

    # Create a template from the query string
    template = jinja_env.from_string(query)

    # Render the template with the provided context
    rendered_query = template.render(**context_dict)

    # Execute the query and return results as DataFrame
    pg = ezr.PG()
    pg.query(rendered_query)
    df = pg.to_dataframe()
    return df



import contextlib
import os
import weakref
from io import StringIO


def _get_duck_connection(file_name, overwrite=False, read_only=False):
    """
    A function to get a connection to the local database
    Args:
        reset: if set to True, will blow away any existing database file
    """
    import ibis

    ibis.options.sql.default_limit = None
    if os.path.isfile(file_name) and overwrite:
        os.unlink(file_name)
    conn = ibis.duckdb.connect(file_name, read_only=read_only)
    return conn


def _get_pg_connection(url=None):
    import ibis

    ibis.options.sql.default_limit = None
    from .postgres import pg_creds_from_env

    if url is None:
        url = pg_creds_from_env()
    conn = ibis.postgres.connect(url=url)
    return conn


@contextlib.contextmanager
def ibis_conn_to_sqlalchemy_conn(ibis_conn):
    engine = ibis_conn.con
    with engine.begin() as conn:
        yield conn


@contextlib.contextmanager
def ibis_postgres_connection(url=None):
    connection = _get_pg_connection(url)
    try:
        proxy_conn = weakref.proxy(connection)
        yield proxy_conn
    finally:
        connection.con.dispose()
        del connection


@contextlib.contextmanager
def ibis_duck_connection(file_name, overwrite=False, read_only=False):
    connection = _get_duck_connection(
        file_name, overwrite=overwrite, read_only=read_only
    )
    try:
        proxy_conn = weakref.proxy(connection)
        yield proxy_conn
    finally:
        connection.con.dispose()
        del connection


def get_sql(expr):
    """
    Returns the SQL for an ibis expression as a string
    """
    import ibis

    buff = StringIO()
    ibis.show_sql(expr, file=buff)
    return buff.getvalue()


def sql_to_frame(conn, sql):
    """
    Returns a pandas dataframe from a SQL query on an ibis connection
    """
    import pandas as pd

    res = conn.raw_sql(sql)
    df = pd.DataFrame(res.fetchall(), columns=res._metadata.keys)
    return df


def get_order_schema_class():
    import ibis
    import ibis.expr.datatypes as dtypes
    from ibis.expr import schema as sch
    import pandas as pd

    class OrderedSchema(ibis.Schema):
        dt = dtypes

        def ordered_apply_to(self, df, strict=True):
            """
            Ensures the dataframe will have columns listed
            in the same order as the schema definition.
            Args:
                strict:  If set to True (the default) ensures that the entire schema
                        of the resulting dataframe has been properly typed.
            """
            # Make sure the input is a dataframe
            if not isinstance(df, pd.DataFrame):
                raise ValueError("ordered_appy_to only defined for dataframes")

            # Get the columns and datatypes for the schema
            cols, types = zip(*self.items())

            # Limit the frame to have only the desired columns and
            # copy it.  This avoids set with copy as well as ensures
            # that the input frame is not mutated
            df = df[list(cols)].copy()

            # Apply the schema to the frame
            self.apply_to(df)

            # If strict comparison is requested
            if strict and not df.empty:
                # Infer the schema of the dataframe
                actual = sch.infer(df)

                # Created a new ibis schema that mirrors this one
                expected = ibis.Schema(dict(self.items()))

                # If the two schemas are not equal, raise an error
                if actual != expected:
                    s = f"\n\nExpected Schema:\n{expected}\n--------\nResulting Schema\n{actual}"
                    raise TypeError(s)

            return df

    return OrderedSchema



from collections import OrderedDict
from itertools import chain
import copy
from textwrap import dedent


class examples:
    """
    A descriptor whose only purpose is to print help text
    """

    def __get__(self, *args, **kwargs):
        print(
            dedent(
                """
                #=================================================================
                #                Example Usage
                #=================================================================

                # Do imports
                import numpy as np
                from scipy.optimize import fmin
                from easier import ParamState

                # Define a model that gives response values in terms of params
                def model(p):
                    return p.a * p.x_train ** p.n

                # Define a cost function for the optimizer to minimize
                def cost(args, p):
                    '''
                    args: a numpy array of parameters that scipy optimizer passes in
                    p: a ParamState object
                    '''

                    # Update paramstate with the latest values from the optimizer
                    p.ingest(args)

                    # Use the paramstate to generate a "fit" based on current params
                    y_fit = model(p)

                    # Compute the errors
                    err = y_fit - p.y_train

                    # Compute and return the cost
                    cost = np.sum(err ** 2)
                    return cost

                # Make some fake data
                x_train = np.linspace(0, 10, 100)
                y_train = -7 * x_train ** 2
                y_train = y_train + .5 * np.random.randn(len(x_train))


                # Create a paramstate with variable names
                p = ParamState('a n')

                # Specify the data you are fitting
                p.given(
                    x_train=x_train,
                    y_train=y_train
                )


                # Get the initial values for params
                x0 = p.array

                # Run the minimizer to get the optimal params
                xf = fmin(cost, x0, args=(p,))

                # Update ParamState with optimal params
                p.ingest(xf)

                # Print the optimized results
                print(p)
            """
            )
        )
        return None


class ParamState(object):
    """
    See the ParamState.examples attribute for examples
    """

    # This is the default initial value to use for variables.
    INITIAL_VALUE = 1.0

    examples = examples()

    def __init__(self, *args, **kwargs):
        """
        *args: a list of variable names
        **kwargs: any initializations to set initial arg values
        """
        from astropy.units.quantity import Quantity

        # allow variables to be passed in a single string
        if len(args) == 1 and isinstance(args[0], str):
            args = args[0].replace(",", " ").split()

        # this dict will hold all variables
        self.vars = OrderedDict()

        # this keeps track of which variables are fixed
        self._fixed_vars = set()

        # a dict to hold the units for the args
        self.unit_dict = {}

        for arg in chain(args, kwargs.keys()):
            if arg not in self.vars:
                val = kwargs.get(arg, self.INITIAL_VALUE)
                try:
                    if isinstance(val, Quantity):
                        self.unit_dict[arg] = val.unit
                except NameError:
                    pass

                self.add(arg, val)

    def add(self, arg, val=None):
        if val is None:
            val = self.INITIAL_VALUE
        self.vars[arg] = val
        setattr(self, arg, val)

    def __setattr__(self, name, value):
        if hasattr(self, "vars") and name in self.vars:
            self.vars[name] = value
        super(ParamState, self).__setattr__(name, value)

    def clone(self):
        return copy.deepcopy(self)

    def drop(self, *variable_names):
        self._fixed_vars = self._fixed_vars - set(variable_names)
        for v in variable_names:
            if v in self.vars:
                del self.vars[v]
            if v in self.unit_dict:
                del self.unit_dict[v]

    def as_dict(self, copy=False):
        if copy:
            return OrderedDict(**self.vars)
        else:
            return self.vars

    def to_dict(self, copy=False):
        return self.as_dict(copy=copy)

    def __setitem__(self, key, value):
        self.__setattr__(key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def keys(self):
        return self.vars.keys()

    def values(self):
        return self.vars.values()

    def items(self):
        return self.vars.items()

    def ingest(self, array):
        """
        Update the parameter state with values obtained from array
        """
        variables = [k for k in self.vars.keys() if k not in self._fixed_vars]
        if len(variables) != len(array):
            raise ValueError(
                "Array to ingest should have length {}".format(len(variables))
            )
        updates = dict(zip(variables, array))

        # attatch units if needed
        for var, val in updates.items():
            updates[var] = self.unit_dict.get(var, 1) * val

        self.vars.update(updates)
        self.__dict__.update(updates)

    def given(self, **kwargs):
        """
        Supply kwargs that will set constant variables.  These can either
        be new variables or already defined variables.  This allows you
        to easily turn on and off variables to optimize.
        """
        self.vars.update(kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)

        self._fixed_vars = self._fixed_vars.union(set(kwargs.keys()))
        return self

    @property
    def given_params(self):
        return OrderedDict(
            [(k, v) for (k, v) in self.vars.items() if k in self._fixed_vars]
        )

    @property
    def free_params(self):
        return OrderedDict(
            [(k, v) for (k, v) in self.vars.items() if k not in self._fixed_vars]
        )

    def __str__(self):
        """
        Prints a nice representation of the results
        """
        return self.df.to_string()

    def _repr_html_(self):
        # makes for nice display in notebook
        return self.df.to_html()

    def __repr__(self):
        return self.__str__()

    @property
    def df(self):
        # again, import here because don't want user to have to remember import
        import pandas as pd

        rec_list = []
        for k, v in self.vars.items():
            kind = "*" if k in self._fixed_vars else ""
            rec_list.append((k, v, kind))
        df = pd.DataFrame(rec_list, columns=["var", "val", "const"])
        df = df.sort_values(by=["const", "var"])
        df = df.set_index("var")
        df.index.name = None
        return df

    @property
    def array(self):
        """
        Returns only the non-constant variables as an array.  Use this for
        creating the initial state
        """
        # import here so user doesn't have to
        import numpy as np
        from astropy.units.quantity import Quantity

        out = []
        for key, val in self.vars.items():
            if key not in self._fixed_vars:
                try:
                    if isinstance(val, Quantity):
                        val = val.value
                except NameError:
                    pass
                out.append(val)

        return np.array(out)

    @property
    def tuple(self):
        """
        Return a tuple of all variables.  Use this for calling functions
        """
        return tuple(self.vars.values())



import os
import textwrap
import contextlib


@contextlib.contextmanager
def dataset_connection(file_name):
    import dataset

    conn = dataset.connect(f"sqlite:///{file_name}")
    try:
        yield conn
    finally:
        conn.close()


class Tables:
    """
    This descriptor provides attribute acces to the tables in the db.
    Think of it like the .objects attribute on a Django model
    """

    def __init__(self, **kwargs):
        self._table_names = []
        for (
            k,
            v,
        ) in kwargs.items():
            self._table_names.append(k)
            setattr(self, k, v)

    def __str__(self):  # pragma: no cover
        return f"Tables({self._table_names})"

    def __repr__(self):  # pragma: no cover
        return self.__str__()


class MiniTable:
    """
    A wrapper around a dataset table that knows how to dump to
    a pandas dataframe
    """

    def __init__(self, file_name, table_name):
        self._table_name = table_name
        self._file_name = file_name

    def __str__(self):  # pragma: no cover
        return f"MiniTable({self._table_name})"

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    @property
    def df(self):
        import pandas as pd

        with dataset_connection(self._file_name) as connection:
            table = connection[self._table_name]
            df = pd.DataFrame(table.all())
        return df


class table_getter:
    def __get__(self, obj, cls):
        content = {}
        for table_name in obj.table_names:
            content[table_name] = MiniTable(obj.file_name, table_name)
        return Tables(**content)


class MiniModel:
    """
    A class for storing pandas dataframes in sqlite database
    """

    # This is a descriptor that provides attribute-style lookup for tables
    # Think of it like the .objects attribute on django models
    tables = table_getter()

    example = textwrap.dedent(
        """
        import pandas as pd
        import easier as ezr

        # Creat dataframe to store
        df_one = pd.DataFrame(
            [
                {'a': 1, 'b': pd.Timestamp('1/1/2022')},
                {'a': 2, 'b': pd.Timestamp('1/2/2022')}
            ]
        )

        # Create second dataframe to store
        df_two = df_one.copy()
        df_two.loc[0, 'a'] = 7
        df_two.loc[0, 'b'] = pd.Timestamp('11/18/2022')


        # Create a sqlite file managed by minimodel
        mm = ezr.MiniModel('rob.sqlite', overwrite=True)

        # Create a table from the first dataframe
        mm.create('table_one', df_one)

        # Insert additional records to the first table
        mm.insert('table_one', df_two)

        # Create a second table from the second dataframe
        # (Note I'm creating the table using insert, which
        # is also allowed)
        mm.insert('table_two', df_two)

        # Get the saved table as a dataframe
        df1 = mm.tables.table_one.df
        df2 = mm.tables.table_two.df
        print()
        print(df1.to_string())
        print()
        print(df2.to_string())

        # Create a third dataframe to demonstrate upserting
        dfup = df1.copy().drop('id', axis=1)
        dfup.loc[:, 'b'] = pd.Timestamp('11/18/2050')

        # Upsert into table two
        df3_pre = mm.tables.table_two.df
        mm.upsert('table_two', ['a'], dfup)
        df3_post = mm.tables.table_two.df
        print()
        print(df3_pre.to_string())
        print()
        print(df3_post.to_string())
    """
    )

    def __init__(
        self, file_name="./mini_model.sqlite", overwrite=False, read_only=False
    ):
        """
        Args:
            file_name: The name of the sqlite file
            overwrite: Blow away any existing file with that name and ovewrite
            read_only: Prevents unintentional overwriting of a database
        """
        # Store the absolute path to the file and handle overwriting
        self.file_name = os.path.realpath(os.path.expanduser(file_name))
        self._read_only = read_only
        if overwrite and os.path.isfile(file_name):
            os.unlink(file_name)

    def __str__(self):  # pragma: no cover
        return f"MiniModel({os.path.basename(self.file_name)})"

    def __repr__(self):  # pragma: no cover
        return self.__str__()

    # @property
    # def connection(self):
    #     # Returns a dataset connection to the db
    #     # Look at api for dataset connections.  Tables are accessed like dictionary
    #     # lookups on the conection
    #     import dataset
    #     conn = dataset.connect(f'sqlite:///{self.file_name}')
    #     return conn

    @property
    def table_names(self):
        """
        A utility property that returns all the table names in the db
        """
        names = []
        if os.path.isfile(self.file_name):
            with dataset_connection(self.file_name) as connection:
                names = connection.inspect.get_table_names()
        return names

    def _framify(self, data):
        """
        A utility method to turn inputs into dataframe.
        Inputs can be dicts, series or dataframes.

        If dicts or series, they are transformed into rows of a dataframe
        """
        import pandas as pd

        if isinstance(data, dict):
            data = pd.Series(data)
        if isinstance(data, pd.Series):
            data = pd.DataFrame(data).T

        if not isinstance(data, pd.DataFrame):
            raise ValueError("data must be a dict, pandas series or pandas dataframe")
        return data

    def _listify(self, df):
        """
        Utility method to turn a dataframe into a list of dicts that dataset
        knows how to insert into the db
        """
        import pandas as pd

        datetime_cols = [
            c for c in df.columns if isinstance(df[c].iloc[0], pd.Timestamp)
        ]
        recs = df.to_dict("records")
        for col in datetime_cols:
            for rec in recs:
                rec[col] = rec[col].to_pydatetime()
        return recs

    def _ensure_writeable(self):
        if self._read_only:
            raise ValueError("Trying to write to read-only db")

    def insert(self, table_name, data):
        """
        Method for inserting data (series, dict or dataframe) into db
        """
        self._ensure_writeable()

        # Turn the input into a dataframe
        df = self._framify(data)

        # Turn the dataframe into a record list
        recs = self._listify(df)

        # Insert the record list
        with dataset_connection(self.file_name) as connection:
            connection[table_name].insert_many(recs)
        return self

    def create(self, table_name, data):
        """
        Creates (i.e. overwrites) a table in the database with the supplied data
        """
        self._ensure_writeable()

        # Drop the table if it exists
        if table_name in self.table_names:
            with dataset_connection(self.file_name) as connection:
                connection[table_name].drop()

        # Insert the new records
        self.insert(table_name, data)
        return self

    def upsert(self, table_name, keys, data):
        """
        Upsert records into specified table.
        Args:
            table_name: the name of the table to upsert
                  keys: a list of column names that the input data must match
                        in the db for a record to be updated in stead of inserted
                  data: The data to upsert.  Single records can be series or dicts.
                        Multiple records defined using dataframes

        """
        self._ensure_writeable()

        # Turn the data into a dataframe
        df = self._framify(data)

        # Turn a dataframe into lists
        recs = self._listify(df)

        # Do the upsert
        with dataset_connection(self.file_name) as connection:
            connection[table_name].upsert_many(recs, keys)
        return self

    def query(self, sql):
        """
        Run a SQL query against the database
        """
        import pandas as pd

        with dataset_connection(self.file_name) as connection:
            df = pd.DataFrame(connection.query(sql))
        return df



class VonMisesFitter:
    def __init__(self, mod_value):
        """
        A fitter for periodic data using a von mises basis.
        Args:
            mod_value: This value mods to zero to make the data periodic
        """
        self.mod_value = mod_value
        self.model = None

    def _to_angle(self, x):
        """
        A utility to convert x to an angle between 0 and 2pi
        """
        import numpy as np

        x = np.array(x)
        return np.mod(x, self.mod_value) * 2 * np.pi / self.mod_value

    def fit(self, x, y, points=10, sigma_bin_factor=2, regularizer=None):
        """
        Fits the data to the von mises basis.
        Args:
            x: The x data
            y: The y data
            points: The number of points to use in computing the basis
            sigma_bin_factor: The standard deviation is the bin width times this factor
            regularizer: The regularizer to use in the fit.  If None, no regularizer is used
        """
        return self._fit(x, y, points, sigma_bin_factor, regularizer, predict=False)

    def fit_predict(self, x, y, points=10, sigma_bin_factor=2, regularizer=None):
        """
        Same as fit, but returns the prediction on the training x values
        """
        return self._fit(x, y, points, sigma_bin_factor, regularizer, predict=True)

    def _fit(
        self, x, y, points=10, sigma_bin_factor=2, regularizer=None, predict=False
    ):
        """
        A utility that does that actual fitting/predicting
        """
        import numpy as np
        from sklearn.linear_model import Ridge, LinearRegression

        self.points = points
        self.sigma_bin_factor = sigma_bin_factor

        x = np.array(x)
        y = np.array(y)
        X = self.get_design_matrix(x, points, sigma_bin_factor)

        if regularizer is None:
            self.model = LinearRegression(fit_intercept=False)
        else:
            self.model = Ridge(alpha=regularizer, fit_intercept=False)

        self.model.fit(X, y)
        if predict:
            return self.model.predict(X)

    def predict(self, x):
        """
        Predicts the y values for the given x values"""
        import numpy as np

        if self.model is None:
            raise ValueError("You must run .fit() before running .predict()")
        x = np.array(x)
        X = self.get_design_matrix(x, self.points, self.sigma_bin_factor)
        return self.model.predict(X)

    def get_design_matrix(self, x, points, sigma_bin_factor):
        """
        Given the x data and the number of points to use for constructing the basis, and the
        sigma_bin_factor, this function returns the design matrix.
        Args:
            x: The x data
            points: The number of points to use in computing the basis
            sigma_bin_factor: The standard deviation is the bin width times this factor
        """
        import numpy as np

        if points < 2:
            raise ValueError("Can only specifiy points > 1")

        from scipy.stats import vonmises

        x = np.array(x)
        theta = self._to_angle(x)

        points = int(points)
        points = np.linspace(0, 2 * np.pi, points + 1)[:points]

        delta = points[1] - points[0]
        sigma = sigma_bin_factor * delta
        kappa = 1 / (sigma**2)

        A = np.zeros((len(x), len(points)))
        for col, point in enumerate(points):
            dist = vonmises(kappa, loc=point)
            A[:, col] = dist.pdf(theta)
        return A



import itertools
import json
import os
from textwrap import dedent
import string
from typing import Optional
from .utils import cached_property

import pandas as pd
import numpy as np
import datetime

# Setting the config file name like this is ugly.  Need a better way.
config_file_locations = [
    os.path.expanduser("/developer/.config/gspread/service_account.json"),
    os.path.expanduser("~/.config/gspread/service_account.json"),
]
for CONFIG_FILE_NAME in config_file_locations:
    if os.path.isfile(CONFIG_FILE_NAME):
        break


class Email:
    def __init__(self, config_file_name):
        self.config_file_name = config_file_name

    @cached_property
    def email(self):
        try:
            with open(self.config_file_name) as buff:
                blob = json.load(buff)

            return blob["client_email"]
        except:  # noqa
            return "email_not_found"

    def __get__(self, obj, objclass):
        return self.email

    def __set__(self, obj, val):
        raise NotImplementedError("Can't set the email property")


class Example:
    def __get__(self, obj, objclass):
        example = dedent(
            f"""
            Share document with this email address:

                {objclass.email}

            Then:
            goog = ezr.GSheet(document_name, sheet_name)
            goog.store_frame(edf)

        """
        )
        return example

    def __set__(self, obj, val):
        raise NotImplementedError("Can't set the example property")


class GSheet:
    email = Email(CONFIG_FILE_NAME)
    example = Example()

    def __init__(self, doc, sheet):
        """
        Expects service account config at ~/.config/gspread/service_account.json

        See http://gspread.readthedocs.io/en/latest/oauth2.html for setting
        up sheets to enable api.
        """
        self._api = None
        self._document = None
        self._sheet = None
        self._df_cache = {}
        self.document = self.api.open(doc)
        self.sheet = self.document.worksheet(sheet)

        if not os.path.isfile(CONFIG_FILE_NAME):
            url = "http://gspread.readthedocs.io/en/latest/oauth2.html"
            msg = f"\n\nFile does not exist:\n{CONFIG_FILE_NAME}\n\nSee {url}"
            raise RuntimeError(msg)

    @classmethod
    def print_example(cls):
        print(cls.example)

    @cached_property
    def col_to_num(self):
        """
        Translate a spreadsheet column name to a column number
        """
        uppers = list(string.ascii_uppercase)
        return {
            val: ind + 1
            for (ind, val) in enumerate(
                "".join(list(t)) for t in list(itertools.product([""] + uppers, uppers))
            )
        }

    @cached_property
    def num_to_col(self):
        """
        Translate a column number to a spreadsheet column name
        """
        return {num: col for (col, num) in self.col_to_num.items()}

    @property
    def api(self):
        # Import here to avoid gspread dependency
        import gspread

        gc = gspread.service_account(CONFIG_FILE_NAME)
        return gc

    def to_dataframe_as_values(self):
        lol = self.sheet.get_all_values()
        return pd.DataFrame(lol)

    def to_dataframe(self, header_row=1, reload=True):
        """
        Dump the current sheet to a dataframe using the specified row
        as a source for column names
        """
        # get a key to identify this frame in the cache
        key = (self.document.title, self.sheet.title)

        # fill cache if needed and returned cached frame
        df = self._df_cache.get(key)
        if df is None or reload:
            recs = self.sheet.get_all_records(head=header_row)
            columns = self.sheet.row_values(header_row)
            self._df_cache[key] = pd.DataFrame(recs)[columns]
        return self._df_cache[key].copy()

    def read_cell(self, coord):
        cell = self.sheet.acell(coord, value_render_option="UNFORMATTED_VALUE")
        return cell.value

    def write_cell(self, coord, value):
        self.write_cells([(coord, value)])

    def read_formula(self, coord):
        cell = self.sheet.acell(coord, value_render_option="FORMULA")
        return cell.value

    def write_formula(self, coord, value):
        self.sheet.update_acell(coord, value)

    def write_cells(self, tuples):
        """
        Tuples of (coord, value)
        where coord is something like 'C15'
        """
        cells = []
        for coord, val in tuples:
            cell = self.sheet.acell(coord)
            cell.value = val
            cells.append(cell)

        self.sheet.update_cells(cells)

    def store_frame_to_coords(
        self,
        df: pd.DataFrame,
        top_left_coord: str,
        clear_to_bottom: bool = False,
        max_num_rows: Optional[int] = None,
        max_num_cols: Optional[int] = None,
    ):
        """
        Args:
            df: the dataframe to store
            top_left_coord: The spreadsheet coordinate (e.g. 'A15') of the top left corner
            max_num_rows: Limit the dataframe to have this many rows before pushing
            max_num_cols: Limit the dataframe to have this mancy cols before pushing
        """
        top_left_coord = top_left_coord.upper()

        df = df.reset_index(drop=True).fillna("")

        if max_num_rows:
            df = df.iloc[:max_num_rows, :]
        if max_num_cols:
            df = df.iloc[:, :max_num_cols]

        cell = self.sheet.acell(top_left_coord)
        top_row = cell.row
        left_col = cell.col

        bottom_row = cell.row + len(df)
        right_col = cell.col + len(df.columns) - 1

        if clear_to_bottom:
            range_string = (
                f"'{self.sheet.title}'!{top_left_coord}:{self.num_to_col[right_col]}"
            )
            self.document.values_clear(range_string)
        else:
            range_string = f"'{self.sheet.title}'!{top_left_coord}:{self.num_to_col[right_col]}{bottom_row}"
            self.document.values_clear(range_string)

        cells = self.sheet.range(top_row, left_col, bottom_row, right_col)

        gsheet_epoc = datetime.datetime(1899, 12, 30)
        for cell in cells:
            frame_row = cell.row - top_row - 1
            frame_col = cell.col - left_col
            if cell.row == top_row:
                cell.value = list(df.columns)[frame_col]
            else:
                value = df.values[frame_row, frame_col]
                # This is a hack.  Dates will be exported as days from epoch
                # Format this as date to get the dates you want
                if isinstance(value, pd.Timestamp):
                    value = (value.to_pydatetime() - gsheet_epoc).days
                elif value == "":
                    value = None
                elif type(value) in [np.int64, np.int32, np.int16, np.int8]:
                    value = int(value)
                cell.value = value
        self.sheet.update_cells(cells)

    def store_frame(self, df, starting_row=1, total_rows=None, total_cols=None):
        """
        Clears the specified sheet and repopulates from specified dataframe
        """
        df = df.reset_index(drop=True).fillna("")
        self.sheet.clear()

        if total_rows is None:
            needed_row_count = len(df) + 20
        else:
            needed_row_count = total_rows

        if total_cols is None:
            needed_col_count = len(df.columns) + 20
        else:
            needed_col_count = total_cols

        self.sheet.resize(rows=needed_row_count, cols=needed_col_count)

        min_row = starting_row
        max_row = min_row + len(df)
        min_col = 1
        max_col = len(df.columns)
        cells = self.sheet.range(min_row, min_col, max_row, max_col)

        gsheet_epoc = datetime.datetime(1899, 12, 30)
        for cell in cells:
            if cell.row == starting_row:
                cell.value = list(df.columns)[cell.col - 1]
            else:
                value = df.values[cell.row - starting_row - 1, cell.col - 1]
                # This is a hack.  Dates will be exported as days from epoch
                # Format this as date to get the dates you want
                if isinstance(value, pd.Timestamp):
                    value = (value.to_pydatetime() - gsheet_epoc).days
                elif value == "":
                    value = None
                elif type(value) in [np.int64, np.int32, np.int16, np.int8]:
                    value = int(value)
                cell.value = value
        self.sheet.update_cells(cells)



def mem_get():
    import psutil

    py_proc = psutil.Process()
    return int(py_proc.memory_full_info().uss / float(2**20))


def mem_show(tag=""):
    print("__memory__: {}M  {}".format(mem_get(), tag))



# import multiprocessing to avoid this bug (http://bugs.python.org/issue15881#msg170215)
import multiprocessing

assert multiprocessing
import re
from setuptools import setup, find_packages


def get_version():
    """
    Extracts the version number from the version.py file.
    """
    VERSION_FILE = "easier/version.py"
    mo = re.search(
        r'^__version__ = [\'"]([^\'"]*)[\'"]', open(VERSION_FILE, "rt").read(), re.M
    )
    if mo:
        return mo.group(1)
    else:
        raise RuntimeError("Unable to find version string in {0}.".format(VERSION_FILE))


install_requires = ["click"]

tests_require = [
    "coverage",
    "flake8",
    "pytest",
]

extras_require = {"dev": tests_require}

setup(
    name="easier",
    version=get_version(),
    description="Tools for analysis",
    long_description="Tools for analysis",
    url="https://github.com/robdmc/easier",
    author="Rob deCarvalho",
    author_email="unlisted@unlisted.net",
    keywords="",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    license="MIT",
    include_package_data=True,
    test_suite="nose.collector",
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require=extras_require,
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "ezr.gsheet = easier.cli:gsheet",
            "ezr.gsheet_push = easier.cli:gsheet_push",
            "ezr.sfdc = easier.cli:sfdc",
            "ezr.debug_notebook = easier.cli:nb_debug",
        ],
    },
)



from collections import Counter
from .shaper import Shaper


def sigma_edit_series(in_series, sigma_thresh, iter_counter=None, max_iter=20):
    """
    Workhorse recursive sigma edit function.
    """
    import numpy as np

    iter_counter = Counter() if iter_counter is None else iter_counter

    if in_series.count() == 0:
        msg = "Error:  No non-NaN values from which to remove outliers"
        raise ValueError(msg)

    iter_counter.update("n")
    if iter_counter["n"] > max_iter:
        msg = "Error:  Max Number of iterations exceeded in sigma-editing"
        raise ValueError(msg)

    resid = in_series - in_series.mean()
    std = resid.std()
    sigma_t = sigma_thresh * std
    outside = resid.abs() >= sigma_t
    if any(outside):
        in_series.loc[outside] = np.nan
        in_series = sigma_edit_series(in_series, sigma_thresh, iter_counter, max_iter)

    return in_series


def kill_outliers_sigma_edit(data, sigma_thresh=3, max_iter=20):
    """
    Recursive sigma edit setting outliers to NaN.
    For 2-d arrays and dataframes, all elements will be
    included in the sampe.

    Args:
        data: ndarray, series or dataframe
        sigma_thresh: the threshold standard dev for editing
        max_iter: The maximum iterations allowed.
    """
    shaper = Shaper()
    x = shaper.flatten(data)
    x = sigma_edit_series(x, sigma_thresh=sigma_thresh, max_iter=max_iter)
    return shaper.expand(x)


def kill_outliers_iqr(data, multiple=1.5):
    """
    Identify outliers using the IQR method and null outliers
    to NaN.  For 2-d nd-arrays and dataframes all elements
    are used in the sample.

    Args:
        data: dataframe, series or ndarray
        multiple: the iqr multiple to use in outlier removal
    """
    import numpy as np

    shaper = Shaper()
    x = shaper.flatten(data)

    # Run outlier detection
    q1, q2 = tuple(np.percentile(x, [25, 75]))
    iqr = q2 - q1
    lower = q1 - multiple * iqr
    upper = q2 + multiple * iqr
    out = x.where((lower <= x) & (x <= upper))
    return shaper.expand(out)
